open BinPos
open Coqlib
open Datatypes
open Iteration
open Lattice
open ListSet
open Maps
open MetatheoryAtom

module type DATAFLOW_SOLVER = 
 sig 
  module L : 
   SEMILATTICE
  
  val fixpoint :
    AtomImpl.atom list ATree.t -> (AtomImpl.atom -> L.t -> L.t) ->
    (AtomImpl.atom*L.t) list -> L.t AMap.t option
 end

module type NODE_SET = 
 sig 
  type t 
  
  val add : AtomImpl.atom -> t -> t
  
  val pick : t -> (AtomImpl.atom*t) option
  
  val initial : AtomImpl.atom list ATree.t -> t
 end

module Dataflow_Solver : 
 functor (LAT:SEMILATTICE) ->
 functor (NS:NODE_SET) ->
 DATAFLOW_SOLVER with module L = LAT

module type BACKWARD_DATAFLOW_SOLVER = 
 sig 
  module L : 
   SEMILATTICE
  
  val fixpoint :
    AtomImpl.atom list ATree.t -> (AtomImpl.atom -> L.t -> L.t) ->
    (AtomImpl.atom*L.t) list -> L.t AMap.t option
 end

module Backward_Dataflow_Solver : 
 functor (LAT:SEMILATTICE) ->
 functor (NS:NODE_SET) ->
 BACKWARD_DATAFLOW_SOLVER with module L = LAT

module type ORDERED_TYPE_WITH_TOP = 
 sig 
  type t 
  
  val top : t
 end

module type BBLOCK_SOLVER = 
 sig 
  module L : 
   ORDERED_TYPE_WITH_TOP
  
  val fixpoint :
    AtomImpl.atom list ATree.t -> (AtomImpl.atom -> L.t -> L.t) ->
    AtomImpl.atom -> L.t AMap.t option
 end

module BBlock_solver : 
 functor (LAT:ORDERED_TYPE_WITH_TOP) ->
 BBLOCK_SOLVER with module L = LAT

module AtomNodeSet : 
 sig 
  type t = AtomImpl.atom set
  
  val add : AtomImpl.atom -> t -> t
  
  val pick : t -> (AtomImpl.atom*AtomImpl.atom set) option
  
  val initial : AtomImpl.atom list ATree.t -> t
 end

module Dataflow_Solver_Var_Top : 
 functor (NS:NODE_SET) ->
 sig 
  module L : 
   sig 
    type t = AtomImpl.atom set option
    
    val eq_dec : t -> t -> bool
    
    val beq : t -> t -> bool
    
    val top : t
    
    val bot : t
    
    val lub : t -> t -> t
    
    val add : t -> AtomImpl.atom -> t
    
    val lubs : t list -> t
   end
  
  type state = { st_in : L.t AMap.t; st_wrk : NS.t }
  
  val state_rect : (L.t AMap.t -> NS.t -> 'a1) -> state -> 'a1
  
  val state_rec : (L.t AMap.t -> NS.t -> 'a1) -> state -> 'a1
  
  val st_in : state -> L.t AMap.t
  
  val st_wrk : state -> NS.t
  
  val start_state_in : (AtomImpl.atom*L.t) list -> L.t AMap.t
  
  val start_state :
    AtomImpl.atom list ATree.t -> (AtomImpl.atom*L.t) list -> state
  
  val propagate_succ : state -> L.t -> AtomImpl.atom -> state
  
  val propagate_succ_list : state -> L.t -> AtomImpl.atom set -> state
  
  val step :
    AtomImpl.atom list ATree.t -> (AtomImpl.atom -> L.t -> L.t) -> state ->
    (L.t AMap.t, state) sum
  
  val fixpoint :
    AtomImpl.atom list ATree.t -> (AtomImpl.atom -> L.t -> L.t) ->
    (AtomImpl.atom*L.t) list -> positive -> L.t AMap.t option
 end

