open Bool
open Coqlib
open FSetInterface
open List0
open ListSet
open Maps
open MetatheoryAtom

module type SEMILATTICE = 
 sig 
  type t 
  
  val beq : t -> t -> bool
  
  val bot : t
  
  val lub : t -> t -> t
 end

module type SEMILATTICE_WITH_TOP = 
 sig 
  type t 
  
  val beq : t -> t -> bool
  
  val bot : t
  
  val lub : t -> t -> t
  
  val top : t
 end

module LPMap = 
 functor (L:SEMILATTICE_WITH_TOP) ->
 struct 
  type t_ =
    | Bot_except of L.t ATree.t
    | Top_except of L.t ATree.t
  
  (** val t__rect :
      (L.t ATree.t -> 'a1) -> (L.t ATree.t -> 'a1) -> t_ -> 'a1 **)
  
  let t__rect f f0 = function
    | Bot_except x -> f x
    | Top_except x -> f0 x
  
  (** val t__rec :
      (L.t ATree.t -> 'a1) -> (L.t ATree.t -> 'a1) -> t_ -> 'a1 **)
  
  let t__rec f f0 = function
    | Bot_except x -> f x
    | Top_except x -> f0 x
  
  type t = t_
  
  (** val get : AtomImpl.atom -> t -> L.t **)
  
  let get p = function
    | Bot_except m ->
        (match ATree.get p m with
           | Some x0 -> x0
           | None -> L.bot)
    | Top_except m ->
        (match ATree.get p m with
           | Some x0 -> x0
           | None -> L.top)
  
  (** val set : AtomImpl.atom -> L.t -> t -> t **)
  
  let set p v = function
    | Bot_except m -> Bot_except
        (if L.beq v L.bot then ATree.remove p m else ATree.set p v m)
    | Top_except m -> Top_except
        (if L.beq v L.top then ATree.remove p m else ATree.set p v m)
  
  (** val beq : t -> t -> bool **)
  
  let beq x y =
    match x with
      | Bot_except m ->
          (match y with
             | Bot_except n -> ATree.beq L.beq m n
             | Top_except t0 -> false)
      | Top_except m ->
          (match y with
             | Bot_except t0 -> false
             | Top_except n -> ATree.beq L.beq m n)
  
  (** val bot : t_ **)
  
  let bot =
    Bot_except ATree.empty
  
  (** val top : t_ **)
  
  let top =
    Top_except ATree.empty
  
  (** val opt_lub : L.t -> L.t -> L.t option **)
  
  let opt_lub x y =
    let z = L.lub x y in if L.beq z L.top then None else Some z
  
  (** val lub : t -> t -> t **)
  
  let lub x y =
    match x with
      | Bot_except m ->
          (match y with
             | Bot_except n -> Bot_except
                 (ATree.combine (fun a b ->
                   match a with
                     | Some u ->
                         (match b with
                            | Some v -> Some (L.lub u v)
                            | None -> a)
                     | None -> b) m n)
             | Top_except n -> Top_except
                 (ATree.combine (fun a b ->
                   match a with
                     | Some u ->
                         (match b with
                            | Some v -> opt_lub u v
                            | None -> None)
                     | None -> b) m n))
      | Top_except m ->
          (match y with
             | Bot_except n -> Top_except
                 (ATree.combine (fun a b ->
                   match a with
                     | Some u ->
                         (match b with
                            | Some v -> opt_lub u v
                            | None -> a)
                     | None -> None) m n)
             | Top_except n -> Top_except
                 (ATree.combine (fun a b ->
                   match a with
                     | Some u ->
                         (match b with
                            | Some v -> opt_lub u v
                            | None -> None)
                     | None -> None) m n))
 end

module LFSet = 
 functor (S:S) ->
 struct 
  type t = S.t
  
  (** val beq : t -> t -> bool **)
  
  let beq =
    S.equal
  
  (** val bot : t **)
  
  let bot =
    S.empty
  
  (** val lub : t -> t -> t **)
  
  let lub =
    S.union
 end

module LFlat = 
 functor (X:EQUALITY_TYPE) ->
 struct 
  type t_ =
    | Bot
    | Inj of X.t
    | Top
  
  (** val t__rect : 'a1 -> (X.t -> 'a1) -> 'a1 -> t_ -> 'a1 **)
  
  let t__rect f f0 f1 = function
    | Bot -> f
    | Inj x -> f0 x
    | Top -> f1
  
  (** val t__rec : 'a1 -> (X.t -> 'a1) -> 'a1 -> t_ -> 'a1 **)
  
  let t__rec f f0 f1 = function
    | Bot -> f
    | Inj x -> f0 x
    | Top -> f1
  
  type t = t_
  
  (** val beq : t -> t -> bool **)
  
  let beq x y =
    match x with
      | Bot -> (match y with
                  | Bot -> true
                  | _ -> false)
      | Inj u ->
          (match y with
             | Inj v -> if X.eq u v then true else false
             | _ -> false)
      | Top -> (match y with
                  | Top -> true
                  | _ -> false)
  
  (** val bot : t **)
  
  let bot =
    Bot
  
  (** val top : t **)
  
  let top =
    Top
  
  (** val lub : t -> t -> t **)
  
  let lub x y =
    match x with
      | Bot -> y
      | Inj a ->
          (match y with
             | Bot -> x
             | Inj b -> if X.eq a b then Inj a else Top
             | Top -> Top)
      | Top -> (match y with
                  | Bot -> x
                  | _ -> Top)
 end

module LBoolean = 
 struct 
  type t = bool
  
  (** val beq : t -> t -> bool **)
  
  let beq =
    eqb
  
  (** val bot : bool **)
  
  let bot =
    false
  
  (** val top : bool **)
  
  let top =
    true
  
  (** val lub : t -> t -> bool **)
  
  let lub x y =
    if x then true else y
 end

module Dominators = 
 struct 
  type t = AtomImpl.atom set option
  
  (** val eq_dec : t -> t -> bool **)
  
  let eq_dec x y =
    match x with
      | Some s ->
          (match y with
             | Some s0 -> AtomSet.set_eq_dec AtomImpl.eq_atom_dec s s0
             | None -> false)
      | None -> (match y with
                   | Some s -> false
                   | None -> true)
  
  (** val beq : t -> t -> bool **)
  
  let beq x y =
    if eq_dec x y then true else false
  
  (** val top : t **)
  
  let top =
    Some empty_set
  
  (** val bot : t **)
  
  let bot =
    None
  
  (** val lub : t -> t -> t **)
  
  let lub x y =
    match x with
      | Some cx ->
          (match y with
             | Some cy -> Some (set_inter AtomImpl.eq_atom_dec cx cy)
             | None -> x)
      | None -> y
  
  (** val add : t -> AtomImpl.atom -> t **)
  
  let add x a =
    match x with
      | Some cx -> Some (a::cx)
      | None -> None
  
  (** val lubs : t list -> t **)
  
  let lubs pds =
    fold_left (fun acc p -> lub acc p) pds bot
 end

