open Bool
open Coqlib
open FSetInterface
open List0
open ListSet
open Maps
open MetatheoryAtom

module type SEMILATTICE = 
 sig 
  type t 
  
  val beq : t -> t -> bool
  
  val bot : t
  
  val lub : t -> t -> t
 end

module type SEMILATTICE_WITH_TOP = 
 sig 
  type t 
  
  val beq : t -> t -> bool
  
  val bot : t
  
  val lub : t -> t -> t
  
  val top : t
 end

module LPMap : 
 functor (L:SEMILATTICE_WITH_TOP) ->
 sig 
  type t_ =
    | Bot_except of L.t ATree.t
    | Top_except of L.t ATree.t
  
  val t__rect : (L.t ATree.t -> 'a1) -> (L.t ATree.t -> 'a1) -> t_ -> 'a1
  
  val t__rec : (L.t ATree.t -> 'a1) -> (L.t ATree.t -> 'a1) -> t_ -> 'a1
  
  type t = t_
  
  val get : AtomImpl.atom -> t -> L.t
  
  val set : AtomImpl.atom -> L.t -> t -> t
  
  val beq : t -> t -> bool
  
  val bot : t_
  
  val top : t_
  
  val opt_lub : L.t -> L.t -> L.t option
  
  val lub : t -> t -> t
 end

module LFSet : 
 functor (S:S) ->
 sig 
  type t = S.t
  
  val beq : t -> t -> bool
  
  val bot : t
  
  val lub : t -> t -> t
 end

module LFlat : 
 functor (X:EQUALITY_TYPE) ->
 sig 
  type t_ =
    | Bot
    | Inj of X.t
    | Top
  
  val t__rect : 'a1 -> (X.t -> 'a1) -> 'a1 -> t_ -> 'a1
  
  val t__rec : 'a1 -> (X.t -> 'a1) -> 'a1 -> t_ -> 'a1
  
  type t = t_
  
  val beq : t -> t -> bool
  
  val bot : t
  
  val top : t
  
  val lub : t -> t -> t
 end

module LBoolean : 
 sig 
  type t = bool
  
  val beq : t -> t -> bool
  
  val bot : bool
  
  val top : bool
  
  val lub : t -> t -> bool
 end

module Dominators : 
 sig 
  type t = AtomImpl.atom set option
  
  val eq_dec : t -> t -> bool
  
  val beq : t -> t -> bool
  
  val top : t
  
  val bot : t
  
  val lub : t -> t -> t
  
  val add : t -> AtomImpl.atom -> t
  
  val lubs : t list -> t
 end

