open BinPos
open Coqlib
open Datatypes
open Integers
open Maps

module OrderedPositive = 
 struct 
  type t = positive
  
  (** val compare : t -> t -> positive OrderedType.coq_Compare **)
  
  let compare x y =
    if plt x y
    then OrderedType.LT
    else if peq x y then OrderedType.EQ else OrderedType.GT
  
  (** val eq_dec : t -> t -> bool **)
  
  let eq_dec =
    peq
 end

module OrderedInt = 
 struct 
  (** val wordsize_one : nat **)
  
  let wordsize_one = O
  
  type t = Int.int
  
  (** val compare : t -> t -> t OrderedType.coq_Compare **)
  
  let compare x y =
    let s = zlt (Int.unsigned wordsize_one x) (Int.unsigned wordsize_one y)
    in
    if s
    then OrderedType.LT
    else let s0 = Int.eq_dec wordsize_one x y in
         if s0 then OrderedType.EQ else OrderedType.GT
  
  (** val eq_dec : t -> t -> bool **)
  
  let eq_dec =
    Int.eq_dec wordsize_one
 end

module OrderedIndexed = 
 functor (A:INDEXED_TYPE) ->
 struct 
  type t = A.t
  
  (** val compare : t -> t -> t OrderedType.coq_Compare **)
  
  let compare x y =
    match OrderedPositive.compare (A.index x) (A.index y) with
      | OrderedType.LT -> OrderedType.LT
      | OrderedType.EQ -> OrderedType.EQ
      | OrderedType.GT -> OrderedType.GT
  
  (** val eq_dec : t -> t -> bool **)
  
  let eq_dec x y =
    peq (A.index x) (A.index y)
 end

module OrderedPair = 
 functor (A:OrderedType.OrderedType) ->
 functor (B:OrderedType.OrderedType) ->
 struct 
  type t = A.t*B.t
  
  (** val compare : t -> t -> t OrderedType.coq_Compare **)
  
  let compare x y =
    match A.compare (fst x) (fst y) with
      | OrderedType.LT -> OrderedType.LT
      | OrderedType.EQ ->
          (match B.compare (snd x) (snd y) with
             | OrderedType.LT -> OrderedType.LT
             | OrderedType.EQ -> OrderedType.EQ
             | OrderedType.GT -> OrderedType.GT)
      | OrderedType.GT -> OrderedType.GT
  
  (** val eq_dec : t -> t -> bool **)
  
  let eq_dec x y =
    if A.eq_dec (fst x) (fst y) then B.eq_dec (snd x) (snd y) else false
 end

