open Datatypes
open Kildall
open Lattice
open List0
open Maps
open MetatheoryAtom
open Dom_set
open Infrastructure
open Syntax

module ReachDS = Dataflow_Solver(LBoolean)(AtomNodeSet)

(** val areachable_aux : LLVMsyntax.fdef -> bool AMap.t option **)

let areachable_aux f =
  match LLVMinfra.getEntryBlock f with
    | Some b ->
        let LLVMsyntax.Coq_block_intro (le, phinodes5, cmds5, terminator5) =
          b
        in
        ReachDS.fixpoint (Cfg.successors f) (fun pc r -> r) ((le,true)::[])
    | None -> None

(** val areachable : LLVMsyntax.fdef -> bool AMap.t **)

let areachable f =
  match areachable_aux f with
    | Some rs -> rs
    | None -> AMap.init true

(** val get_reachable_labels :
    LLVMsyntax.l list -> bool AMap.t -> LLVMsyntax.l list -> LLVMsyntax.l
    list **)

let rec get_reachable_labels bd rd acc =
  match bd with
    | [] -> acc
    | l0::bd' ->
        if AMap.get l0 rd
        then get_reachable_labels bd' rd (l0::acc)
        else get_reachable_labels bd' rd acc

(** val reachablity_analysis :
    LLVMsyntax.fdef -> LLVMsyntax.l list option **)

let reachablity_analysis f =
  match LLVMinfra.getEntryBlock f with
    | Some b ->
        (match areachable_aux f with
           | Some rd -> Some (get_reachable_labels (Cfg.bound_fdef f) rd [])
           | None -> None)
    | None -> None

(** val cmds_dominates_cmd :
    LLVMsyntax.cmds -> LLVMsyntax.id -> AtomImpl.atom list **)

let rec cmds_dominates_cmd cs id0 =
  match cs with
    | [] -> []
    | c1::cs' ->
        let ctx = cmds_dominates_cmd cs' id0 in
        if AtomImpl.eq_atom_dec (LLVMinfra.getCmdLoc c1) id0
        then []
        else (match LLVMinfra.getCmdID c1 with
                | Some id1 -> id1::ctx
                | None -> ctx)

(** val inscope_of_block :
    LLVMsyntax.fdef -> LLVMsyntax.l -> AtomImpl.atom list option ->
    LLVMsyntax.l -> AtomImpl.atom list option **)

let inscope_of_block f l1 opt_ctx lbl =
  match opt_ctx with
    | Some ctx ->
        (match LLVMinfra.lookupBlockViaLabelFromFdef f lbl with
           | Some b ->
               if AtomImpl.eq_atom_dec lbl l1
               then Some ctx
               else Some (app (LLVMinfra.getBlockIDs b) ctx)
           | None -> None)
    | None -> None

(** val init_scope :
    LLVMsyntax.fdef -> LLVMsyntax.phinodes -> LLVMsyntax.cmds ->
    LLVMsyntax.id -> AtomImpl.atom list **)

let init_scope f ps cs id0 =
  if in_dec LLVMinfra.id_dec id0 (LLVMinfra.getPhiNodesIDs ps)
  then LLVMinfra.getArgsIDsOfFdef f
  else app (LLVMinfra.getPhiNodesIDs ps)
         (app (cmds_dominates_cmd cs id0) (LLVMinfra.getArgsIDsOfFdef f))

(** val inscope_of_id :
    LLVMsyntax.fdef -> LLVMsyntax.block -> LLVMsyntax.id -> AtomImpl.atom
    list option **)

let inscope_of_id f b1 id0 =
  let LLVMsyntax.Coq_block_intro (l1, ps, cs, terminator5) = b1 in
  fold_left (inscope_of_block f l1) (AlgDom.sdom f l1) (Some
    (init_scope f ps cs id0))

(** val inscope_of_cmd :
    LLVMsyntax.fdef -> LLVMsyntax.block -> LLVMsyntax.cmd -> AtomImpl.atom
    list option **)

let inscope_of_cmd f b1 c =
  let id0 = LLVMinfra.getCmdLoc c in inscope_of_id f b1 id0

(** val inscope_of_tmn :
    LLVMsyntax.fdef -> LLVMsyntax.block -> LLVMsyntax.terminator ->
    AtomImpl.atom list option **)

let inscope_of_tmn f b1 tmn =
  let LLVMsyntax.Coq_block_intro (l1, ps, cs, terminator5) = b1 in
  fold_left (inscope_of_block f l1) (AlgDom.sdom f l1) (Some
    (app (LLVMinfra.getPhiNodesIDs ps)
      (app (LLVMinfra.getCmdsIDs cs) (LLVMinfra.getArgsIDsOfFdef f))))

(** val defs_dominate :
    LLVMsyntax.fdef -> LLVMsyntax.block -> LLVMsyntax.block ->
    LLVMsyntax.insn -> AtomImpl.atom list option **)

let defs_dominate f curb incomingb = function
  | LLVMsyntax.Coq_insn_phinode p ->
      let LLVMsyntax.Coq_block_intro (l5, phinodes5, cmds5, tmn) = incomingb
      in
      inscope_of_tmn f incomingb tmn
  | LLVMsyntax.Coq_insn_cmd c -> inscope_of_cmd f curb c
  | LLVMsyntax.Coq_insn_terminator tmn -> inscope_of_tmn f curb tmn

