open BinPos
open Datatypes
open List0
open ListSet
open Maps
open MetatheoryAtom
open Infrastructure
open Syntax

module Cfg = 
 functor (T:TREE) ->
 struct 
  module XTree = More_Tree_Properties(T)
 end

module ACfg = Cfg(ATree)

module PCfg = Cfg(PTree)

(** val successors_blocks : LLVMsyntax.blocks -> LLVMsyntax.ls ATree.t **)

let rec successors_blocks = function
  | [] -> ATree.empty
  | b::bs' ->
      let LLVMsyntax.Coq_block_intro (l0, phinodes5, cmds5, tmn) = b in
      ATree.set l0 (LLVMinfra.successors_terminator tmn)
        (successors_blocks bs')

(** val successors : LLVMsyntax.fdef -> LLVMsyntax.ls ATree.t **)

let successors = function
  | LLVMsyntax.Coq_fdef_intro (fheader5, bs) -> successors_blocks bs

(** val bound_blocks : LLVMsyntax.blocks -> AtomImpl.atom set **)

let rec bound_blocks = function
  | [] -> empty_set
  | b::bs' ->
      let LLVMsyntax.Coq_block_intro (l0, phinodes5, cmds5, tmn) = b in
      l0::(bound_blocks bs')

(** val bound_fdef : LLVMsyntax.fdef -> AtomImpl.atom set **)

let bound_fdef = function
  | LLVMsyntax.Coq_fdef_intro (fheader5, bs) -> bound_blocks bs

(** val predecessors : LLVMsyntax.fdef -> LLVMsyntax.ls ATree.t **)

let predecessors f =
  XATree.make_predecessors (successors f)

(** val has_no_predecessors : LLVMsyntax.fdef -> LLVMsyntax.block -> bool **)

let has_no_predecessors f b =
  match XATree.successors_list (predecessors f) (LLVMinfra.getBlockLabel b) with
    | [] -> true
    | l::l0 -> false

(** val coq_Pcubeplus : positive -> positive **)

let coq_Pcubeplus = Camlcoq.pcubeplus

(** val plength : 'a1 list -> positive **)

let plength ls1 =
  coq_P_of_succ_nat (length ls1)

(** val pnum_of_blocks_in_fdef : LLVMsyntax.fdef -> positive **)

let pnum_of_blocks_in_fdef = function
  | LLVMsyntax.Coq_fdef_intro (fheader5, bs) ->
      fold_left (fun acc x -> coq_Psucc acc) bs (Coq_xI Coq_xH)

(** val num_iters : LLVMsyntax.fdef -> positive **)

let num_iters f =
  let pn = pnum_of_blocks_in_fdef f in coq_Pcubeplus pn

