open BinPos
open Datatypes
open List0
open ListSet
open Maps
open MetatheoryAtom
open Infrastructure
open Syntax

module Cfg : 
 functor (T:TREE) ->
 sig 
  module XTree : 
   sig 
    module TProp : 
     sig 
      val f' : ('a2 -> T.elt -> 'a1 -> 'a2) -> 'a2 -> (T.elt*'a1) -> 'a2
     end
    
    val successors_list : 'a1 list T.t -> T.elt -> 'a1 list
    
    val add_successors :
      T.elt list T.t -> T.elt -> T.elt list -> T.elt list T.t
    
    val make_predecessors : T.elt list T.t -> T.elt list T.t
    
    val children_of_tree : 'a1 list T.t -> 'a1 list
    
    val parents_of_tree : 'a1 T.t -> T.elt list
    
    val elements_of_cfg :
      T.elt list T.t -> (T.elt -> T.elt -> bool) -> T.elt list
   end
 end

module ACfg : 
 sig 
  module XTree : 
   sig 
    module TProp : 
     sig 
      val f' :
        ('a2 -> ATree.elt -> 'a1 -> 'a2) -> 'a2 -> (ATree.elt*'a1) -> 'a2
     end
    
    val successors_list : 'a1 list ATree.t -> ATree.elt -> 'a1 list
    
    val add_successors :
      ATree.elt list ATree.t -> ATree.elt -> ATree.elt list -> ATree.elt list
      ATree.t
    
    val make_predecessors : ATree.elt list ATree.t -> ATree.elt list ATree.t
    
    val children_of_tree : 'a1 list ATree.t -> 'a1 list
    
    val parents_of_tree : 'a1 ATree.t -> ATree.elt list
    
    val elements_of_cfg :
      ATree.elt list ATree.t -> (ATree.elt -> ATree.elt -> bool) -> ATree.elt
      list
   end
 end

module PCfg : 
 sig 
  module XTree : 
   sig 
    module TProp : 
     sig 
      val f' :
        ('a2 -> PTree.elt -> 'a1 -> 'a2) -> 'a2 -> (PTree.elt*'a1) -> 'a2
     end
    
    val successors_list : 'a1 list PTree.t -> PTree.elt -> 'a1 list
    
    val add_successors :
      PTree.elt list PTree.t -> PTree.elt -> PTree.elt list -> PTree.elt list
      PTree.t
    
    val make_predecessors : PTree.elt list PTree.t -> PTree.elt list PTree.t
    
    val children_of_tree : 'a1 list PTree.t -> 'a1 list
    
    val parents_of_tree : 'a1 PTree.t -> PTree.elt list
    
    val elements_of_cfg :
      PTree.elt list PTree.t -> (PTree.elt -> PTree.elt -> bool) -> PTree.elt
      list
   end
 end

val successors_blocks : LLVMsyntax.blocks -> LLVMsyntax.ls ATree.t

val successors : LLVMsyntax.fdef -> LLVMsyntax.ls ATree.t

val bound_blocks : LLVMsyntax.blocks -> AtomImpl.atom set

val bound_fdef : LLVMsyntax.fdef -> AtomImpl.atom set

val predecessors : LLVMsyntax.fdef -> LLVMsyntax.ls ATree.t

val has_no_predecessors : LLVMsyntax.fdef -> LLVMsyntax.block -> bool

val coq_Pcubeplus : positive -> positive

val plength : 'a1 list -> positive

val pnum_of_blocks_in_fdef : LLVMsyntax.fdef -> positive

val num_iters : LLVMsyntax.fdef -> positive

