open BinPos
open Datatypes
open Maps
open Specif
open Syntax

type __ = Obj.t
let __ = let rec f _ = Obj.repr f in Obj.repr f

type coq_Frame = { coq_Fr_name : LLVMsyntax.l; coq_Fr_scs : LLVMsyntax.ls }

(** val coq_Frame_rect :
    (LLVMsyntax.l -> LLVMsyntax.ls -> 'a1) -> coq_Frame -> 'a1 **)

let coq_Frame_rect f f0 =
  let { coq_Fr_name = x; coq_Fr_scs = x0 } = f0 in f x x0

(** val coq_Frame_rec :
    (LLVMsyntax.l -> LLVMsyntax.ls -> 'a1) -> coq_Frame -> 'a1 **)

let coq_Frame_rec f f0 =
  let { coq_Fr_name = x; coq_Fr_scs = x0 } = f0 in f x x0

(** val coq_Fr_name : coq_Frame -> LLVMsyntax.l **)

let coq_Fr_name x = x.coq_Fr_name

(** val coq_Fr_scs : coq_Frame -> LLVMsyntax.ls **)

let coq_Fr_scs x = x.coq_Fr_scs

(** val find_next_visit_in_scs :
    unit ATree.t -> LLVMsyntax.ls -> (LLVMsyntax.l*LLVMsyntax.ls) option **)

let rec find_next_visit_in_scs visited = function
  | [] -> None
  | sc::scs' ->
      (match ATree.get sc visited with
         | Some u -> find_next_visit_in_scs visited scs'
         | None -> Some (sc,scs'))

type coq_PostOrder = { coq_PO_cnt : positive; coq_PO_a2p : positive ATree.t }

(** val coq_PostOrder_rect :
    (positive -> positive ATree.t -> 'a1) -> coq_PostOrder -> 'a1 **)

let coq_PostOrder_rect f p =
  let { coq_PO_cnt = x; coq_PO_a2p = x0 } = p in f x x0

(** val coq_PostOrder_rec :
    (positive -> positive ATree.t -> 'a1) -> coq_PostOrder -> 'a1 **)

let coq_PostOrder_rec f p =
  let { coq_PO_cnt = x; coq_PO_a2p = x0 } = p in f x x0

(** val coq_PO_cnt : coq_PostOrder -> positive **)

let coq_PO_cnt x = x.coq_PO_cnt

(** val coq_PO_a2p : coq_PostOrder -> positive ATree.t **)

let coq_PO_a2p x = x.coq_PO_a2p

(** val postorder_inc : coq_PostOrder -> LLVMsyntax.l -> coq_PostOrder **)

let postorder_inc po v =
  let { coq_PO_cnt = cnt; coq_PO_a2p = a2p } = po in
  { coq_PO_cnt = (coq_Psucc cnt); coq_PO_a2p = (ATree.set v cnt a2p) }

(** val find_next_visit_in_stk :
    LLVMsyntax.ls ATree.t -> unit ATree.t -> coq_PostOrder -> coq_Frame list
    -> ((LLVMsyntax.l*coq_Frame list)*coq_PostOrder, coq_PostOrder) sum **)

let rec find_next_visit_in_stk succs visited po = function
  | [] -> Coq_inr po
  | f::frs' ->
      let { coq_Fr_name = v; coq_Fr_scs = scs } = f in
      (match find_next_visit_in_scs visited scs with
         | Some p ->
             let sc,scs' = p in
             Coq_inl ((sc,({ coq_Fr_name = sc; coq_Fr_scs =
             (XATree.successors_list succs sc) }::({ coq_Fr_name = v;
             coq_Fr_scs = scs' }::frs'))),po)
         | None ->
             find_next_visit_in_stk succs visited (postorder_inc po v) frs')

(** val dfs_aux1_func :
    (LLVMsyntax.ls ATree.t, (unit ATree.t, (coq_PostOrder, (coq_Frame list,
    __) sigT) sigT) sigT) sigT -> coq_PostOrder **)

let rec dfs_aux1_func x =
  let succs = projT1 x in
  let visited = projT1 (projT2 x) in
  let po = projT1 (projT2 (projT2 x)) in
  let stk = projT1 (projT2 (projT2 (projT2 x))) in
  let dfs_aux2 = fun succs0 visited0 po0 stk0 ->
    let y = Coq_existT (succs0, (Coq_existT (visited0, (Coq_existT (po0,
      (Coq_existT (stk0, __)))))))
    in
    dfs_aux1_func y
  in
  let filtered_var = find_next_visit_in_stk succs visited po stk in
  let branch_0 = fun po' -> po' in
  let branch_1 = fun next stk' po' ->
    dfs_aux2 succs (ATree.set next () visited) po' stk'
  in
  (match filtered_var with
     | Coq_inl p ->
         let p0,po' = p in let next,stk' = p0 in branch_1 next stk' po'
     | Coq_inr po' -> branch_0 po')

(** val dfs_aux1 :
    LLVMsyntax.ls ATree.t -> unit ATree.t -> coq_PostOrder -> coq_Frame list
    -> coq_PostOrder **)

let dfs_aux1 succs visited po stk =
  dfs_aux1_func (Coq_existT (succs, (Coq_existT (visited, (Coq_existT (po,
    (Coq_existT (stk, __))))))))

(** val iter : nat -> ('a1 -> 'a1) -> 'a1 -> 'a1 **)

let rec iter n f g =
  match n with
    | O -> g
    | S p -> f (iter p f g)

type dfs_aux_type =
  LLVMsyntax.ls ATree.t -> unit ATree.t -> coq_PostOrder -> coq_Frame list ->
  coq_PostOrder

(** val dfs_aux_F :
    dfs_aux_type -> LLVMsyntax.ls ATree.t -> unit ATree.t -> coq_PostOrder ->
    coq_Frame list -> coq_PostOrder **)

let dfs_aux_F f succs visited po stk =
  match find_next_visit_in_stk succs visited po stk with
    | Coq_inl p ->
        let p0,po' = p in
        let next,stk' = p0 in f succs (ATree.set next () visited) po' stk'
    | Coq_inr po' -> po'

(** val dfs_aux_terminates_func_obligation_4 :
    LLVMsyntax.ls ATree.t -> unit ATree.t -> coq_PostOrder -> coq_Frame list
    -> (LLVMsyntax.ls ATree.t -> unit ATree.t -> coq_PostOrder -> coq_Frame
    list -> __ -> __ -> coq_PostOrder) -> LLVMsyntax.l -> coq_Frame list ->
    coq_PostOrder -> coq_PostOrder **)

let dfs_aux_terminates_func_obligation_4 succs visited po stk dfs_aux_terminates0 next stk' po' =
  dfs_aux_terminates0 succs (ATree.set next () visited) po' stk' __ __

(** val dfs_aux_terminates_func :
    (LLVMsyntax.ls ATree.t, (unit ATree.t, (coq_PostOrder, (coq_Frame list,
    __) sigT) sigT) sigT) sigT -> coq_PostOrder **)

let rec dfs_aux_terminates_func x =
  let succs = projT1 x in
  let visited = projT1 (projT2 x) in
  let po = projT1 (projT2 (projT2 x)) in
  let stk = projT1 (projT2 (projT2 (projT2 x))) in
  let dfs_aux_terminates0 = fun succs0 visited0 po0 stk0 ->
    let y = Coq_existT (succs0, (Coq_existT (visited0, (Coq_existT (po0,
      (Coq_existT (stk0, __)))))))
    in
    dfs_aux_terminates_func y
  in
  let filtered_var = find_next_visit_in_stk succs visited po stk in
  let branch_0 = fun po' -> po' in
  let branch_1 =
    dfs_aux_terminates_func_obligation_4 succs visited po stk
      (fun succs0 visited0 po0 stk0 _ _ ->
      dfs_aux_terminates0 succs0 visited0 po0 stk0)
  in
  (match filtered_var with
     | Coq_inl p ->
         let p0,po' = p in let next,stk' = p0 in branch_1 next stk' po'
     | Coq_inr po' -> branch_0 po')

(** val dfs_aux_terminates :
    LLVMsyntax.ls ATree.t -> unit ATree.t -> coq_PostOrder -> coq_Frame list
    -> coq_PostOrder **)

let dfs_aux_terminates succs visited po stk =
  dfs_aux_terminates_func (Coq_existT (succs, (Coq_existT (visited,
    (Coq_existT (po, (Coq_existT (stk, __))))))))

(** val dfs_aux :
    LLVMsyntax.ls ATree.t -> unit ATree.t -> coq_PostOrder -> coq_Frame list
    -> coq_PostOrder **)

let dfs_aux succs visited po stk =
  dfs_aux_terminates succs visited po stk

(** val dfs :
    LLVMsyntax.ls ATree.t -> LLVMsyntax.l -> positive -> coq_PostOrder **)

let dfs succs entry pinit =
  dfs_aux succs ATree.empty { coq_PO_cnt = pinit; coq_PO_a2p = ATree.empty }
    ({ coq_Fr_name = entry; coq_Fr_scs =
    (XATree.successors_list succs entry) }::[])

module Inv = 
 struct 
  
 end

module Order = 
 struct 
  
 end

module ReachableEntry = 
 struct 
  
 end

module ReachableSucc = 
 struct 
  
 end

module NumberedAreReachable = 
 struct 
  
 end

module Injective = 
 struct 
  
 end

