open BinPos
open Datatypes
open FSetAVL
open FSetFacts
open Int0
open List0
open Maps
open MetatheoryAtom
open Ordered
open Specif
open Syntax

let __ = let rec f _ = Obj.repr f in Obj.repr f

module type PNODE_SET = 
 sig 
  type t 
  
  val add : positive -> t -> t
  
  val pick : t -> (positive*t) option
  
  val initial : positive list PTree.t -> t
 end

module PositiveSet = Make(OrderedPositive)

module PositiveSetFacts = Facts(PositiveSet)

module PNodeSetMax = 
 struct 
  type t = PositiveSet.t
  
  (** val add : positive -> t -> t **)
  
  let add n s =
    PositiveSet.add n s
  
  (** val pick : t -> (PositiveSet.elt*PositiveSet.t) option **)
  
  let pick s =
    match PositiveSet.max_elt s with
      | Some n -> Some (n,(PositiveSet.remove n s))
      | None -> None
  
  (** val initial : positive list PTree.t -> PositiveSet.t **)
  
  let initial successors =
    PTree.fold (fun s pc scs -> PositiveSet.add pc s) successors
      PositiveSet.empty
 end

module PNodeSetMin = 
 struct 
  type t = PositiveSet.t
  
  (** val add : positive -> t -> t **)
  
  let add n s =
    PositiveSet.add n s
  
  (** val pick : t -> (PositiveSet.elt*PositiveSet.t) option **)
  
  let pick s =
    match PositiveSet.min_elt s with
      | Some n -> Some (n,(PositiveSet.remove n s))
      | None -> None
  
  (** val initial : positive list PTree.t -> PositiveSet.t **)
  
  let initial successors =
    PTree.fold (fun s pc scs -> PositiveSet.add pc s) successors
      PositiveSet.empty
 end

module type MERGE = 
 sig 
  val merge : positive list -> positive list -> positive list*bool
 end

module MergeLt = 
 struct 
  (** val merge_aux_func :
      (positive list, (positive list, positive list*bool) sigT) sigT ->
      positive list*bool **)
  
  let rec merge_aux_func x =
    let l1 = projT1 x in
    let l2 = projT1 (projT2 x) in
    let acc = projT2 (projT2 x) in
    let merge_aux0 = fun l3 l4 acc0 ->
      let y = Coq_existT (l3, (Coq_existT (l4, acc0))) in merge_aux_func y
    in
    let branch_0 = fun rl changed ->
      let branch_0 = fun p1 l1' p2 l2' ->
        let filtered_var = coq_Pcompare p1 p2 Eq in
        let branch_0 = fun _ -> merge_aux0 l1' l2' ((p1::rl),changed) in
        let branch_1 = fun _ -> merge_aux0 l1' l2 (rl,true) in
        let branch_2 = fun _ -> merge_aux0 l1 l2' (rl,changed) in
        (match filtered_var with
           | Eq -> branch_0 __
           | Lt -> branch_1 __
           | Gt -> branch_2 __)
      in
      let branch_1 = fun wildcard' -> acc in
      let branch_2 = fun wildcard' wildcard'0 -> rl,true in
      (match l1 with
         | [] -> branch_1 l2
         | wildcard'::wildcard'0 ->
             (match l2 with
                | [] -> branch_2 wildcard' wildcard'0
                | p2::l2' -> branch_0 wildcard' wildcard'0 p2 l2'))
    in
    let rl,changed = acc in branch_0 rl changed
  
  (** val merge_aux :
      positive list -> positive list -> (positive list*bool) -> positive
      list*bool **)
  
  let merge_aux l1 l2 acc =
    merge_aux_func (Coq_existT (l1, (Coq_existT (l2, acc))))
  
  (** val merge : positive list -> positive list -> positive list*bool **)
  
  let merge l1 l2 =
    let rl,changed = merge_aux l1 l2 ([],false) in (rev rl),changed
 end

module type LATTICEELT = 
 sig 
  type t 
  
  val bot : t
  
  val lub : t -> t -> t*bool
  
  val top : t
 end

module LATTICEELT_MAP = 
 functor (L:LATTICEELT) ->
 struct 
  
 end

module Doms = 
 functor (MG:MERGE) ->
 struct 
  type t = positive list option
  
  (** val bot : t **)
  
  let bot =
    None
  
  (** val lub : t -> t -> t*bool **)
  
  let lub x y =
    match x with
      | Some x' ->
          (match y with
             | Some y' -> let r,changed = MG.merge x' y' in (Some r),changed
             | None -> x,false)
      | None -> (match y with
                   | Some l0 -> y,true
                   | None -> x,false)
  
  (** val top : t **)
  
  let top =
    Some []
  
  (** val transfer : positive -> t -> t **)
  
  let transfer n = function
    | Some ps -> Some (n::ps)
    | None -> None
 end

(** val atolist_ptolist_fun :
    positive ATree.t -> positive list -> AtomImpl.atom -> positive list **)

let atolist_ptolist_fun a2p acc ato =
  match ATree.get ato a2p with
    | Some pto -> pto::acc
    | None -> acc

(** val atolist_ptolist :
    positive ATree.t -> LLVMsyntax.l list -> positive list **)

let atolist_ptolist a2p tolist =
  fold_left (atolist_ptolist_fun a2p) tolist []

(** val asucc_psucc :
    positive ATree.t -> positive list PTree.t -> LLVMsyntax.l -> LLVMsyntax.l
    list -> positive list PTree.t **)

let asucc_psucc a2p pred from tolist =
  match ATree.get from a2p with
    | Some pfrom -> PTree.set pfrom (atolist_ptolist a2p tolist) pred
    | None -> pred

(** val asuccs_psuccs :
    positive ATree.t -> LLVMsyntax.ls ATree.t -> positive list PTree.t **)

let asuccs_psuccs mapping succs =
  ATree.fold (asucc_psucc mapping) succs PTree.empty

