open BinPos
open Datatypes
open FSetAVL
open FSetFacts
open Int0
open List0
open Maps
open MetatheoryAtom
open Ordered
open Specif
open Syntax

module type PNODE_SET = 
 sig 
  type t 
  
  val add : positive -> t -> t
  
  val pick : t -> (positive*t) option
  
  val initial : positive list PTree.t -> t
 end

module PositiveSet : 
 sig 
  module X' : 
   sig 
    type t = positive
    
    val eq_dec : positive -> positive -> bool
    
    val compare : positive -> positive -> comparison
   end
  
  module MSet : 
   sig 
    module Raw : 
     sig 
      type elt = positive
      
      type tree =
        | Leaf
        | Node of tree * positive * tree * Z_as_Int.int
      
      type t = tree
      
      val height : t -> Z_as_Int.int
      
      val cardinal : t -> nat
      
      val empty : tree
      
      val is_empty : tree -> bool
      
      val mem : positive -> tree -> bool
      
      val singleton : positive -> tree
      
      val create : tree -> positive -> tree -> tree
      
      val assert_false : tree -> positive -> tree -> tree
      
      val bal : t -> positive -> t -> tree
      
      val add : positive -> tree -> tree
      
      val join : tree -> elt -> t -> t
      
      val remove_min : tree -> elt -> t -> t*elt
      
      val merge : tree -> tree -> tree
      
      val remove : positive -> tree -> tree
      
      val min_elt : tree -> positive option
      
      val max_elt : tree -> positive option
      
      val choose : tree -> positive option
      
      val concat : tree -> tree -> tree
      
      type triple = { t_left : t; t_in : bool; t_right : t }
      
      val t_left : triple -> t
      
      val t_in : triple -> bool
      
      val t_right : triple -> t
      
      val split : positive -> tree -> triple
      
      val inter : tree -> tree -> tree
      
      val diff : tree -> tree -> tree
      
      val union : tree -> tree -> tree
      
      val elements_aux : positive list -> t -> positive list
      
      val elements : t -> positive list
      
      val filter_acc : (elt -> bool) -> tree -> tree -> tree
      
      val filter : (elt -> bool) -> tree -> tree
      
      val partition_acc : (elt -> bool) -> (t*t) -> t -> t*t
      
      val partition : (elt -> bool) -> t -> t*t
      
      val for_all : (elt -> bool) -> tree -> bool
      
      val exists_ : (elt -> bool) -> tree -> bool
      
      val fold : (elt -> 'a1 -> 'a1) -> t -> 'a1 -> 'a1
      
      val subsetl : (t -> bool) -> positive -> tree -> bool
      
      val subsetr : (t -> bool) -> positive -> tree -> bool
      
      val subset : tree -> tree -> bool
      
      type enumeration =
        | End
        | More of elt * t * enumeration
      
      val cons : tree -> enumeration -> enumeration
      
      val compare_more :
        positive -> (enumeration -> comparison) -> enumeration -> comparison
      
      val compare_cont :
        tree -> (enumeration -> comparison) -> enumeration -> comparison
      
      val compare_end : enumeration -> comparison
      
      val compare : tree -> tree -> comparison
      
      val equal : tree -> tree -> bool
      
      val ltb_tree : positive -> tree -> bool
      
      val gtb_tree : positive -> tree -> bool
      
      val isok : tree -> bool
      
      module MX : 
       sig 
        module OrderTac : 
         sig 
          module OTF : 
           sig 
            type t = positive
            
            val compare : positive -> positive -> comparison
            
            val eq_dec : positive -> positive -> bool
           end
          
          module TO : 
           sig 
            type t = positive
            
            val compare : positive -> positive -> comparison
            
            val eq_dec : positive -> positive -> bool
           end
         end
        
        val eq_dec : positive -> positive -> bool
        
        val lt_dec : positive -> positive -> bool
        
        val eqb : positive -> positive -> bool
       end
      
      type coq_R_bal =
        | R_bal_0 of t * positive * t
        | R_bal_1 of t * positive * t * tree * positive * tree * Z_as_Int.int
        | R_bal_2 of t * positive * t * tree * positive * tree * Z_as_Int.int
        | R_bal_3 of t * positive * t * tree * positive * 
           tree * Z_as_Int.int * tree * positive * 
           tree * Z_as_Int.int
        | R_bal_4 of t * positive * t
        | R_bal_5 of t * positive * t * tree * positive * tree * Z_as_Int.int
        | R_bal_6 of t * positive * t * tree * positive * tree * Z_as_Int.int
        | R_bal_7 of t * positive * t * tree * positive * 
           tree * Z_as_Int.int * tree * positive * 
           tree * Z_as_Int.int
        | R_bal_8 of t * positive * t
      
      type coq_R_remove_min =
        | R_remove_min_0 of tree * elt * t
        | R_remove_min_1 of tree * elt * t * tree * 
           positive * tree * Z_as_Int.int * (t*elt) * 
           coq_R_remove_min * t * elt
      
      type coq_R_merge =
        | R_merge_0 of tree * tree
        | R_merge_1 of tree * tree * tree * positive * tree * Z_as_Int.int
        | R_merge_2 of tree * tree * tree * positive * 
           tree * Z_as_Int.int * tree * positive * 
           tree * Z_as_Int.int * t * elt
      
      type coq_R_min_elt =
        | R_min_elt_0 of tree
        | R_min_elt_1 of tree * tree * positive * tree * Z_as_Int.int
        | R_min_elt_2 of tree * tree * positive * tree * 
           Z_as_Int.int * tree * positive * tree * 
           Z_as_Int.int * positive option * coq_R_min_elt
      
      type coq_R_max_elt =
        | R_max_elt_0 of tree
        | R_max_elt_1 of tree * tree * positive * tree * Z_as_Int.int
        | R_max_elt_2 of tree * tree * positive * tree * 
           Z_as_Int.int * tree * positive * tree * 
           Z_as_Int.int * positive option * coq_R_max_elt
      
      type coq_R_concat =
        | R_concat_0 of tree * tree
        | R_concat_1 of tree * tree * tree * positive * tree * Z_as_Int.int
        | R_concat_2 of tree * tree * tree * positive * 
           tree * Z_as_Int.int * tree * positive * 
           tree * Z_as_Int.int * t * elt
      
      type coq_R_inter =
        | R_inter_0 of tree * tree
        | R_inter_1 of tree * tree * tree * positive * tree * Z_as_Int.int
        | R_inter_2 of tree * tree * tree * positive * 
           tree * Z_as_Int.int * tree * positive * 
           tree * Z_as_Int.int * t * bool * t * tree * 
           coq_R_inter * tree * coq_R_inter
        | R_inter_3 of tree * tree * tree * positive * 
           tree * Z_as_Int.int * tree * positive * 
           tree * Z_as_Int.int * t * bool * t * tree * 
           coq_R_inter * tree * coq_R_inter
      
      type coq_R_diff =
        | R_diff_0 of tree * tree
        | R_diff_1 of tree * tree * tree * positive * tree * Z_as_Int.int
        | R_diff_2 of tree * tree * tree * positive * 
           tree * Z_as_Int.int * tree * positive * 
           tree * Z_as_Int.int * t * bool * t * tree * 
           coq_R_diff * tree * coq_R_diff
        | R_diff_3 of tree * tree * tree * positive * 
           tree * Z_as_Int.int * tree * positive * 
           tree * Z_as_Int.int * t * bool * t * tree * 
           coq_R_diff * tree * coq_R_diff
      
      type coq_R_union =
        | R_union_0 of tree * tree
        | R_union_1 of tree * tree * tree * positive * tree * Z_as_Int.int
        | R_union_2 of tree * tree * tree * positive * 
           tree * Z_as_Int.int * tree * positive * 
           tree * Z_as_Int.int * t * bool * t * tree * 
           coq_R_union * tree * coq_R_union
      
      module L : 
       sig 
        module MO : 
         sig 
          module OrderTac : 
           sig 
            module OTF : 
             sig 
              type t = positive
              
              val compare : positive -> positive -> comparison
              
              val eq_dec : positive -> positive -> bool
             end
            
            module TO : 
             sig 
              type t = positive
              
              val compare : positive -> positive -> comparison
              
              val eq_dec : positive -> positive -> bool
             end
           end
          
          val eq_dec : positive -> positive -> bool
          
          val lt_dec : positive -> positive -> bool
          
          val eqb : positive -> positive -> bool
         end
       end
      
      val flatten_e : enumeration -> elt list
     end
    
    module E : 
     sig 
      type t = positive
      
      val compare : positive -> positive -> comparison
      
      val eq_dec : positive -> positive -> bool
     end
    
    type elt = positive
    
    type t_ = Raw.t
      (* singleton inductive, whose constructor was Mkt *)
    
    val this : t_ -> Raw.t
    
    type t = t_
    
    val mem : elt -> t -> bool
    
    val add : elt -> t -> t
    
    val remove : elt -> t -> t
    
    val singleton : elt -> t
    
    val union : t -> t -> t
    
    val inter : t -> t -> t
    
    val diff : t -> t -> t
    
    val equal : t -> t -> bool
    
    val subset : t -> t -> bool
    
    val empty : t
    
    val is_empty : t -> bool
    
    val elements : t -> elt list
    
    val choose : t -> elt option
    
    val fold : (elt -> 'a1 -> 'a1) -> t -> 'a1 -> 'a1
    
    val cardinal : t -> nat
    
    val filter : (elt -> bool) -> t -> t
    
    val for_all : (elt -> bool) -> t -> bool
    
    val exists_ : (elt -> bool) -> t -> bool
    
    val partition : (elt -> bool) -> t -> t*t
    
    val eq_dec : t -> t -> bool
    
    val compare : t -> t -> comparison
    
    val min_elt : t -> elt option
    
    val max_elt : t -> elt option
   end
  
  type elt = positive
  
  type t = MSet.t
  
  val empty : t
  
  val is_empty : t -> bool
  
  val mem : elt -> t -> bool
  
  val add : elt -> t -> t
  
  val singleton : elt -> t
  
  val remove : elt -> t -> t
  
  val union : t -> t -> t
  
  val inter : t -> t -> t
  
  val diff : t -> t -> t
  
  val eq_dec : t -> t -> bool
  
  val equal : t -> t -> bool
  
  val subset : t -> t -> bool
  
  val fold : (elt -> 'a1 -> 'a1) -> t -> 'a1 -> 'a1
  
  val for_all : (elt -> bool) -> t -> bool
  
  val exists_ : (elt -> bool) -> t -> bool
  
  val filter : (elt -> bool) -> t -> t
  
  val partition : (elt -> bool) -> t -> t*t
  
  val cardinal : t -> nat
  
  val elements : t -> elt list
  
  val choose : t -> elt option
  
  module MF : 
   sig 
    val eqb : positive -> positive -> bool
   end
  
  val min_elt : t -> elt option
  
  val max_elt : t -> elt option
  
  val compare : t -> t -> t OrderedType.coq_Compare
  
  module E : 
   sig 
    type t = positive
    
    val compare : positive -> positive -> positive OrderedType.coq_Compare
    
    val eq_dec : positive -> positive -> bool
   end
 end

module PositiveSetFacts : 
 sig 
  val eqb : positive -> positive -> bool
 end

module PNodeSetMax : 
 sig 
  type t = PositiveSet.t
  
  val add : positive -> t -> t
  
  val pick : t -> (PositiveSet.elt*PositiveSet.t) option
  
  val initial : positive list PTree.t -> PositiveSet.t
 end

module PNodeSetMin : 
 sig 
  type t = PositiveSet.t
  
  val add : positive -> t -> t
  
  val pick : t -> (PositiveSet.elt*PositiveSet.t) option
  
  val initial : positive list PTree.t -> PositiveSet.t
 end

module type MERGE = 
 sig 
  val merge : positive list -> positive list -> positive list*bool
 end

module MergeLt : 
 sig 
  val merge_aux_func :
    (positive list, (positive list, positive list*bool) sigT) sigT ->
    positive list*bool
  
  val merge_aux :
    positive list -> positive list -> (positive list*bool) -> positive
    list*bool
  
  val merge : positive list -> positive list -> positive list*bool
 end

module type LATTICEELT = 
 sig 
  type t 
  
  val bot : t
  
  val lub : t -> t -> t*bool
  
  val top : t
 end

module LATTICEELT_MAP : 
 functor (L:LATTICEELT) ->
 sig 
  
 end

module Doms : 
 functor (MG:MERGE) ->
 sig 
  type t = positive list option
  
  val bot : t
  
  val lub : t -> t -> t*bool
  
  val top : t
  
  val transfer : positive -> t -> t
 end

val atolist_ptolist_fun :
  positive ATree.t -> positive list -> AtomImpl.atom -> positive list

val atolist_ptolist : positive ATree.t -> LLVMsyntax.l list -> positive list

val asucc_psucc :
  positive ATree.t -> positive list PTree.t -> LLVMsyntax.l -> LLVMsyntax.l
  list -> positive list PTree.t

val asuccs_psuccs :
  positive ATree.t -> LLVMsyntax.ls ATree.t -> positive list PTree.t

