open BinPos
open List0
open Maps
open MetatheoryAtom
open Dfs
open Dom_libs
open Dom_list_tree
open Dom_tree
open Dom_type
open Infrastructure
open Push_iter
open Syntax

(** val get_reachable_nodes_fun :
    positive ATree.t -> positive list -> LLVMsyntax.l -> positive list **)

let get_reachable_nodes_fun a2p acc a =
  match ATree.get a a2p with
    | Some p -> p::acc
    | None -> acc

(** val get_reachable_nodes :
    LLVMsyntax.l list -> positive ATree.t -> positive list **)

let get_reachable_nodes bd a2p =
  fold_left (get_reachable_nodes_fun a2p) bd []

(** val asuccs : LLVMsyntax.fdef -> LLVMsyntax.ls ATree.t **)

let asuccs f =
  Cfg.successors f

(** val dom_analyze : LLVMsyntax.fdef -> LDoms.t PMap.t*positive ATree.t **)

let dom_analyze f =
  let asuccs0 = Cfg.successors f in
  (match LLVMinfra.getEntryLabel f with
     | Some le ->
         let { coq_PO_cnt = pO_cnt; coq_PO_a2p = a2p } =
           dfs asuccs0 le Coq_xH
         in
         let psuccs = asuccs_psuccs a2p asuccs0 in
         (match ATree.get le a2p with
            | Some pe -> (pdom_analyze psuccs pe (Cfg.num_iters f)),a2p
            | None -> (PMap.init LDoms.top),ATree.empty)
     | None -> (PMap.init LDoms.top),ATree.empty)

(** val a2p_p2a : positive ATree.t -> LLVMsyntax.l PTree.t **)

let a2p_p2a a2p =
  ATree.fold (fun acc from to0 -> PTree.set to0 from acc) a2p PTree.empty

(** val ps2as_fun :
    LLVMsyntax.l PTree.t -> LLVMsyntax.ls -> positive -> LLVMsyntax.l list **)

let ps2as_fun p2a acc p =
  match PTree.get p p2a with
    | Some a -> a::acc
    | None -> acc

(** val ps2as :
    LLVMsyntax.l PTree.t -> positive list -> LLVMsyntax.l list **)

let ps2as p2a ps =
  fold_left (ps2as_fun p2a) ps []

module AlgDom' = 
 struct 
  (** val p2a_dom :
      LLVMsyntax.l PTree.t -> AtomImpl.atom list -> LDoms.t -> AtomImpl.atom
      list **)
  
  let p2a_dom p2a bd = function
    | Some dts2 -> ps2as p2a dts2
    | None -> bd
  
  (** val bound_dom :
      positive ATree.t -> LLVMsyntax.l PTree.t -> AtomImpl.atom list ->
      LDoms.t PMap.t -> AtomImpl.atom -> AtomImpl.atom list **)
  
  let bound_dom a2p p2a bd res l0 =
    match ATree.get l0 a2p with
      | Some p0 -> p2a_dom p2a bd (PMap.get p0 res)
      | None -> bd
  
  (** val sdom : LLVMsyntax.fdef -> AtomImpl.atom -> AtomImpl.atom list **)
  
  let sdom f =
    let dt,a2p = dom_analyze f in
    let b = Cfg.bound_fdef f in
    let p2a = a2p_p2a a2p in bound_dom a2p p2a b dt
  
  (** val create_dom_tree : LLVMsyntax.fdef -> ATree.elt coq_DTree option **)
  
  let create_dom_tree f =
    let dts,a2p = dom_analyze f in
    (match LLVMinfra.getEntryLabel f with
       | Some le ->
           (match ATree.get le a2p with
              | Some pe ->
                  let res = fun p -> PMap.get p dts in
                  let p2a = a2p_p2a a2p in
                  let pdt =
                    create_dtree pe
                      (get_reachable_nodes (Cfg.bound_fdef f) a2p) res
                  in
                  DTreeConv.sdtree_tdtree p2a pdt
              | None -> None)
       | None -> None)
 end

module AlgDomProps' = AlgDom_Properties(AlgDom')

