open BinPos
open Coqlib
open Datatypes
open Kildall
open Lattice
open List0
open ListSet
open Maps
open MetatheoryAtom
open Dom_type
open Infrastructure
open Syntax

module AlgDom = 
 struct 
  (** val transfer : LLVMsyntax.l -> Dominators.t -> Dominators.t **)
  
  let transfer lbl before =
    Dominators.add before lbl
  
  module DomDS = Dataflow_Solver_Var_Top(AtomNodeSet)
  
  module WorklistProps = 
   struct 
    
   end
  
  module DomsInParents = 
   struct 
    (** val transf : LLVMsyntax.l -> Dominators.t -> Dominators.t **)
    
    let transf =
      transfer
    
    (** val top : Dominators.t **)
    
    let top =
      Dominators.top
    
    (** val bot : Dominators.t **)
    
    let bot =
      Dominators.bot
    
    (** val entrypoints : LLVMsyntax.l -> (AtomImpl.atom*DomDS.L.t) list **)
    
    let entrypoints entry =
      (entry,Dominators.top)::[]
   end
  
  module Termination = 
   struct 
    (** val elements_of_cfg :
        AtomImpl.atom list ATree.t -> AtomImpl.atom list **)
    
    let elements_of_cfg successors =
      XATree.elements_of_cfg successors AtomImpl.eq_atom_dec
    
    (** val psize_of_cfg : AtomImpl.atom list ATree.t -> positive **)
    
    let psize_of_cfg successors =
      coq_Pplus (Cfg.plength (elements_of_cfg successors)) (Coq_xO Coq_xH)
    
    (** val psize_of_dom :
        AtomImpl.atom list ATree.t -> DomDS.L.t -> positive **)
    
    let psize_of_dom successors = function
      | Some x ->
          Cfg.plength (AtomSet.elements_of_set AtomImpl.eq_atom_dec x)
      | None -> psize_of_cfg successors
    
    (** val num_of_doms_fun :
        AtomImpl.atom list ATree.t -> DomDS.L.t AMap.t -> positive ->
        AtomImpl.atom -> positive **)
    
    let num_of_doms_fun successors dmap acc a =
      coq_Pplus acc (psize_of_dom successors (AMap.get a dmap))
    
    (** val num_of_doms :
        AtomImpl.atom list ATree.t -> DomDS.L.t AMap.t -> positive **)
    
    let num_of_doms successors dmap =
      fold_left (num_of_doms_fun successors dmap)
        (elements_of_cfg successors) Coq_xH
    
    (** val num_iters : AtomImpl.atom list ATree.t -> positive **)
    
    let num_iters successors =
      Cfg.coq_Pcubeplus (psize_of_cfg successors)
    
    (** val psize_of_worklist : AtomNodeSet.t -> positive **)
    
    let psize_of_worklist wrk =
      coq_Pplus (Cfg.plength wrk) Coq_xH
    
    (** val num_iters_aux :
        AtomImpl.atom list ATree.t -> DomDS.state -> positive **)
    
    let num_iters_aux successors st =
      coq_Pplus
        (coq_Pmult (psize_of_cfg successors)
          (num_of_doms successors (DomDS.st_in st)))
        (psize_of_worklist (DomDS.st_wrk st))
    
    (** val entrypoints : AtomImpl.atom -> (AtomImpl.atom*DomDS.L.t) list **)
    
    let entrypoints entrypoint =
      (entrypoint,DomDS.L.top)::[]
   end
  
  (** val dom_analyze : LLVMsyntax.fdef -> Dominators.t AMap.t **)
  
  let dom_analyze f = match f with
    | LLVMsyntax.Coq_fdef_intro (fheader5, bs) ->
        (match LLVMinfra.getEntryBlock f with
           | Some b ->
               let LLVMsyntax.Coq_block_intro
                 (le, phinodes5, cmds5, terminator5) = b
               in
               (match DomDS.fixpoint (Cfg.successors_blocks bs) transfer
                        ((le,Dominators.top)::[]) (Cfg.num_iters f) with
                  | Some res -> res
                  | None -> AMap.init Dominators.top)
           | None -> AMap.init Dominators.top)
  
  (** val bound_dom :
      AtomImpl.atom set -> Dominators.t -> AtomImpl.atom set **)
  
  let bound_dom bd = function
    | Some dts2 -> dts2
    | None -> bd
  
  (** val sdom : LLVMsyntax.fdef -> AtomImpl.atom -> AtomImpl.atom set **)
  
  let sdom f =
    let dt = dom_analyze f in
    let b = Cfg.bound_fdef f in (fun l0 -> bound_dom b (AMap.get l0 dt))
  
  module EntryDomsOthers = 
   struct 
    (** val predecessors : LLVMsyntax.blocks -> ATree.elt list ATree.t **)
    
    let predecessors bs =
      XATree.make_predecessors (Cfg.successors_blocks bs)
    
    (** val transf : LLVMsyntax.l -> Dominators.t -> Dominators.t **)
    
    let transf =
      transfer
    
    (** val top : Dominators.t **)
    
    let top =
      Dominators.top
    
    (** val bot : Dominators.t **)
    
    let bot =
      Dominators.bot
    
    (** val lub_of_preds :
        LLVMsyntax.blocks -> DomDS.L.t AMap.t -> AtomImpl.atom -> DomDS.L.t **)
    
    let lub_of_preds bs res n =
      Dominators.lubs
        (map (fun p -> transf p (AMap.get p res))
          (XATree.successors_list (predecessors bs) n))
   end
  
  module DomComplete = 
   struct 
    (** val predecessors : LLVMsyntax.blocks -> ATree.elt list ATree.t **)
    
    let predecessors bs =
      XATree.make_predecessors (Cfg.successors_blocks bs)
    
    (** val transf : LLVMsyntax.l -> Dominators.t -> Dominators.t **)
    
    let transf =
      transfer
    
    (** val top : Dominators.t **)
    
    let top =
      Dominators.top
    
    (** val bot : Dominators.t **)
    
    let bot =
      Dominators.bot
   end
  
  module UnreachableDoms = 
   struct 
    (** val predecessors : LLVMsyntax.blocks -> ATree.elt list ATree.t **)
    
    let predecessors bs =
      XATree.make_predecessors (Cfg.successors_blocks bs)
    
    (** val transf : LLVMsyntax.l -> Dominators.t -> Dominators.t **)
    
    let transf =
      transfer
    
    (** val top : Dominators.t **)
    
    let top =
      Dominators.top
    
    (** val bot : Dominators.t **)
    
    let bot =
      Dominators.bot
   end
 end

module AlgDomProps = AlgDom_Properties(AlgDom)

