open Coqlib
open Datatypes
open List0
open ListSet
open MetatheoryAtom
open Analysis
open Dom_set
open Dom_tree
open Infrastructure
open Syntax

(** val dtree_dom : AtomImpl.atom coq_DTree -> AtomSetImpl.t **)

let rec dtree_dom = function
  | DT_node (l0, dts0) ->
      AtomSetImpl.union (AtomSetImpl.singleton l0) (dtrees_dom dts0)

(** val dtrees_dom : AtomImpl.atom coq_DTrees -> AtomSetImpl.t **)

and dtrees_dom = function
  | DT_nil -> AtomSetImpl.empty
  | DT_cons (dt0, dts0) ->
      AtomSetImpl.union (dtree_dom dt0) (dtrees_dom dts0)

(** val gt_sdom :
    (LLVMsyntax.l -> LLVMsyntax.l set) -> LLVMsyntax.l -> LLVMsyntax.l ->
    bool **)

let gt_sdom res l1 l2 =
  proj_sumbool (in_dec LLVMinfra.l_dec l1 (res l2))

(** val find_min :
    (LLVMsyntax.l -> LLVMsyntax.l set) -> LLVMsyntax.l -> LLVMsyntax.l set ->
    LLVMsyntax.l **)

let rec find_min res acc = function
  | [] -> acc
  | l0::dts' ->
      if gt_sdom res acc l0
      then find_min res l0 dts'
      else find_min res acc dts'

(** val insert_sort_sdom_iter :
    (LLVMsyntax.l -> LLVMsyntax.l set) -> LLVMsyntax.l -> LLVMsyntax.l list
    -> LLVMsyntax.l list -> LLVMsyntax.l list **)

let rec insert_sort_sdom_iter res l0 prefix suffix = match suffix with
  | [] -> rev (l0::prefix)
  | l1::suffix' ->
      if gt_sdom res l0 l1
      then app (rev (l0::prefix)) suffix
      else insert_sort_sdom_iter res l0 (l1::prefix) suffix'

(** val insert_sort_sdom :
    (LLVMsyntax.l -> LLVMsyntax.l set) -> LLVMsyntax.l list -> LLVMsyntax.l
    list -> LLVMsyntax.l list **)

let rec insert_sort_sdom res data acc =
  match data with
    | [] -> acc
    | l1::data' ->
        insert_sort_sdom res data' (insert_sort_sdom_iter res l1 [] acc)

(** val sort_sdom :
    (LLVMsyntax.l -> LLVMsyntax.l set) -> LLVMsyntax.l list -> LLVMsyntax.l
    list **)

let sort_sdom res data =
  insert_sort_sdom res data []

(** val remove_redundant : LLVMsyntax.l list -> LLVMsyntax.l list **)

let rec remove_redundant input = match input with
  | [] -> input
  | a::input' ->
      (match input' with
         | [] -> input
         | b::l0 ->
             if LLVMinfra.l_dec a b
             then remove_redundant input'
             else a::(remove_redundant input'))

(** val compute_sdom_chains_aux :
    (LLVMsyntax.l -> LLVMsyntax.l set) -> LLVMsyntax.l list ->
    (LLVMsyntax.l*LLVMsyntax.l list) list -> (LLVMsyntax.l*LLVMsyntax.l list)
    list **)

let rec compute_sdom_chains_aux res bd acc =
  match bd with
    | [] -> acc
    | l0::bd' ->
        compute_sdom_chains_aux res bd'
          ((l0,(remove_redundant (sort_sdom res (l0::(res l0)))))::acc)

(** val compute_sdom_chains :
    (LLVMsyntax.l -> LLVMsyntax.l set) -> LLVMsyntax.l list ->
    (LLVMsyntax.l*LLVMsyntax.l list) list **)

let compute_sdom_chains res rd =
  compute_sdom_chains_aux res rd []

(** val find_idom_aux :
    (LLVMsyntax.l -> LLVMsyntax.l set) -> LLVMsyntax.l -> LLVMsyntax.l set ->
    LLVMsyntax.l option **)

let rec find_idom_aux res acc = function
  | [] -> Some acc
  | l0::dts' ->
      if in_dec LLVMinfra.l_dec l0 (res acc)
      then find_idom_aux res acc dts'
      else if in_dec LLVMinfra.l_dec acc (res l0)
           then find_idom_aux res l0 dts'
           else None

(** val find_idom :
    (LLVMsyntax.l -> LLVMsyntax.l set) -> LLVMsyntax.l -> LLVMsyntax.l option **)

let find_idom res l0 =
  match res l0 with
    | [] -> None
    | l1::dts0 -> find_idom_aux res l1 dts0

(** val compute_idoms :
    (LLVMsyntax.l -> LLVMsyntax.l set) -> LLVMsyntax.l list ->
    (LLVMsyntax.l*LLVMsyntax.l) list -> (LLVMsyntax.l*LLVMsyntax.l) list **)

let rec compute_idoms res rd acc =
  match rd with
    | [] -> acc
    | l0::rd' ->
        (match find_idom res l0 with
           | Some l1 -> compute_idoms res rd' ((l1,l0)::acc)
           | None -> compute_idoms res rd' acc)

(** val create_dtree : LLVMsyntax.fdef -> AtomImpl.atom coq_DTree option **)

let create_dtree f =
  match LLVMinfra.getEntryLabel f with
    | Some root ->
        (match reachablity_analysis f with
           | Some rd ->
               let dt = AlgDom.sdom f in
               let chains = compute_sdom_chains dt rd in
               Some
               (create_dtree_from_chains AtomImpl.eq_atom_dec root chains)
           | None -> None)
    | None -> None

