open Datatypes
open List0
open ListSet
open Maps

type 'a coq_DTree =
  | DT_node of 'a * 'a coq_DTrees
and 'a coq_DTrees =
  | DT_nil
  | DT_cons of 'a coq_DTree * 'a coq_DTrees

(** val coq_DTree_rect :
    ('a1 -> 'a1 coq_DTrees -> 'a2) -> 'a1 coq_DTree -> 'a2 **)

let coq_DTree_rect f = function
  | DT_node (x, x0) -> f x x0

(** val coq_DTree_rec :
    ('a1 -> 'a1 coq_DTrees -> 'a2) -> 'a1 coq_DTree -> 'a2 **)

let coq_DTree_rec f = function
  | DT_node (x, x0) -> f x x0

(** val coq_DTrees_rect :
    'a2 -> ('a1 coq_DTree -> 'a1 coq_DTrees -> 'a2 -> 'a2) -> 'a1 coq_DTrees
    -> 'a2 **)

let rec coq_DTrees_rect f f0 = function
  | DT_nil -> f
  | DT_cons (d0, d1) -> f0 d0 d1 (coq_DTrees_rect f f0 d1)

(** val coq_DTrees_rec :
    'a2 -> ('a1 coq_DTree -> 'a1 coq_DTrees -> 'a2 -> 'a2) -> 'a1 coq_DTrees
    -> 'a2 **)

let rec coq_DTrees_rec f f0 = function
  | DT_nil -> f
  | DT_cons (d0, d1) -> f0 d0 d1 (coq_DTrees_rec f f0 d1)

(** val size_of_dtrees : 'a1 coq_DTrees -> nat **)

let rec size_of_dtrees = function
  | DT_nil -> O
  | DT_cons (d, dts') -> S (size_of_dtrees dts')

(** val list2tree : 'a1 -> 'a1 list -> 'a1 coq_DTree **)

let rec list2tree hd = function
  | [] -> DT_node (hd, DT_nil)
  | hd'::tl' -> DT_node (hd, (DT_cons ((list2tree hd' tl'), DT_nil)))

(** val dtree_insert :
    ('a1 -> 'a1 -> bool) -> 'a1 coq_DTree -> 'a1 -> 'a1 -> 'a1 list -> 'a1
    coq_DTree **)

let dtree_insert hdec =
  let rec dtree_insert0 dt parent child tl =
    let DT_node (l0, dts0) = dt in
    if hdec parent l0
    then DT_node (l0, (dtrees_insert0 dts0 child tl))
    else dt
  and dtrees_insert0 dts child tl =
    match dts with
      | DT_nil -> DT_cons ((list2tree child tl), DT_nil)
      | DT_cons (dt0, dts0) ->
          let DT_node (l0, d) = dt0 in
          if hdec l0 child
          then (match tl with
                  | [] -> dts
                  | child'::tl' -> DT_cons
                      ((dtree_insert0 dt0 child child' tl'), dts0))
          else DT_cons (dt0, (dtrees_insert0 dts0 child tl))
  in dtree_insert0

(** val dtrees_insert :
    ('a1 -> 'a1 -> bool) -> 'a1 coq_DTrees -> 'a1 -> 'a1 list -> 'a1
    coq_DTrees **)

let dtrees_insert hdec =
  let rec dtree_insert0 dt parent child tl =
    let DT_node (l0, dts0) = dt in
    if hdec parent l0
    then DT_node (l0, (dtrees_insert0 dts0 child tl))
    else dt
  and dtrees_insert0 dts child tl =
    match dts with
      | DT_nil -> DT_cons ((list2tree child tl), DT_nil)
      | DT_cons (dt0, dts0) ->
          let DT_node (l0, d) = dt0 in
          if hdec l0 child
          then (match tl with
                  | [] -> dts
                  | child'::tl' -> DT_cons
                      ((dtree_insert0 dt0 child child' tl'), dts0))
          else DT_cons (dt0, (dtrees_insert0 dts0 child tl))
  in dtrees_insert0

(** val create_dtree_from_chain :
    ('a1 -> 'a1 -> bool) -> 'a1 coq_DTree -> 'a1 list -> 'a1 coq_DTree **)

let create_dtree_from_chain hdec dt = function
  | [] -> dt
  | p::l ->
      (match l with
         | [] -> dt
         | ch::chain' -> dtree_insert hdec dt p ch chain')

(** val create_dtree_from_chains :
    ('a1 -> 'a1 -> bool) -> 'a1 -> ('a1*'a1 list) list -> 'a1 coq_DTree **)

let create_dtree_from_chains hdec pe chains =
  fold_left (fun acc elt0 -> create_dtree_from_chain hdec acc (snd elt0))
    chains (DT_node (pe, DT_nil))

(** val in_children_roots :
    ('a1 -> 'a1 -> bool) -> 'a1 -> 'a1 coq_DTrees -> bool **)

let rec in_children_roots hdec child = function
  | DT_nil -> false
  | DT_cons (d, dts') ->
      let DT_node (l0, d0) = d in
      if hdec l0 child then true else in_children_roots hdec child dts'

(** val is_dtree_edge :
    ('a1 -> 'a1 -> bool) -> 'a1 coq_DTree -> 'a1 -> 'a1 -> bool **)

let is_dtree_edge hdec =
  let rec is_dtree_edge0 dt parent child =
    let DT_node (l0, dts0) = dt in
    if hdec parent l0
    then if in_children_roots hdec child dts0
         then true
         else is_dtrees_edge0 dts0 parent child
    else is_dtrees_edge0 dts0 parent child
  and is_dtrees_edge0 dts parent child =
    match dts with
      | DT_nil -> false
      | DT_cons (dt0, dts0) ->
          if is_dtree_edge0 dt0 parent child
          then true
          else is_dtrees_edge0 dts0 parent child
  in is_dtree_edge0

(** val is_dtrees_edge :
    ('a1 -> 'a1 -> bool) -> 'a1 coq_DTrees -> 'a1 -> 'a1 -> bool **)

let is_dtrees_edge hdec =
  let rec is_dtree_edge0 dt parent child =
    let DT_node (l0, dts0) = dt in
    if hdec parent l0
    then if in_children_roots hdec child dts0
         then true
         else is_dtrees_edge0 dts0 parent child
    else is_dtrees_edge0 dts0 parent child
  and is_dtrees_edge0 dts parent child =
    match dts with
      | DT_nil -> false
      | DT_cons (dt0, dts0) ->
          if is_dtree_edge0 dt0 parent child
          then true
          else is_dtrees_edge0 dts0 parent child
  in is_dtrees_edge0

(** val dtrees_rec2 :
    ('a1 -> 'a1 coq_DTrees -> 'a3 -> 'a2) -> 'a3 -> ('a1 coq_DTree -> 'a2 ->
    'a1 coq_DTrees -> 'a3 -> 'a3) -> 'a1 coq_DTrees -> 'a3 **)

let dtrees_rec2 f f0 f1 =
  let rec f2 = function
    | DT_node (a, d0) -> f a d0 (f3 d0)
  and f3 = function
    | DT_nil -> f0
    | DT_cons (d0, d1) -> f1 d0 (f2 d0) d1 (f3 d1)
  in f3

(** val dtree_rec2 :
    ('a1 -> 'a1 coq_DTrees -> 'a3 -> 'a2) -> 'a3 -> ('a1 coq_DTree -> 'a2 ->
    'a1 coq_DTrees -> 'a3 -> 'a3) -> 'a1 coq_DTree -> 'a2 **)

let dtree_rec2 f f0 f1 =
  let rec f2 = function
    | DT_node (a, d0) -> f a d0 (f3 d0)
  and f3 = function
    | DT_nil -> f0
    | DT_cons (d0, d1) -> f1 d0 (f2 d0) d1 (f3 d1)
  in f2

(** val dtree_mutrec :
    ('a1 -> 'a1 coq_DTrees -> 'a3 -> 'a2) -> 'a3 -> ('a1 coq_DTree -> 'a2 ->
    'a1 coq_DTrees -> 'a3 -> 'a3) -> ('a1 coq_DTree -> 'a2)*('a1 coq_DTrees
    -> 'a3) **)

let dtree_mutrec h1 h2 h3 =
  (dtree_rec2 h1 h2 h3),(dtrees_rec2 h1 h2 h3)

(** val get_dtree_root : 'a1 coq_DTree -> 'a1 **)

let get_dtree_root = function
  | DT_node (l0, d) -> l0

(** val dtree_dom : ('a1 -> 'a1 -> bool) -> 'a1 coq_DTree -> 'a1 set **)

let dtree_dom hdec =
  let rec dtree_dom0 = function
    | DT_node (l0, dts0) -> set_add hdec l0 (dtrees_dom0 dts0)
  and dtrees_dom0 = function
    | DT_nil -> empty_set
    | DT_cons (dt0, dts0) ->
        set_union hdec (dtree_dom0 dt0) (dtrees_dom0 dts0)
  in dtree_dom0

(** val dtrees_dom : ('a1 -> 'a1 -> bool) -> 'a1 coq_DTrees -> 'a1 set **)

let dtrees_dom hdec =
  let rec dtree_dom0 = function
    | DT_node (l0, dts0) -> set_add hdec l0 (dtrees_dom0 dts0)
  and dtrees_dom0 = function
    | DT_nil -> empty_set
    | DT_cons (dt0, dts0) ->
        set_union hdec (dtree_dom0 dt0) (dtrees_dom0 dts0)
  in dtrees_dom0

(** val get_dtrees_roots :
    ('a1 -> 'a1 -> bool) -> 'a1 coq_DTrees -> 'a1 set **)

let rec get_dtrees_roots hdec = function
  | DT_nil -> empty_set
  | DT_cons (dt', dts') ->
      set_add hdec (get_dtree_root dt') (get_dtrees_roots hdec dts')

module DTreeProps = 
 functor (T:TREE) ->
 struct 
  module TCfg = Cfg.Cfg(T)
  
  module XTree = More_Tree_Properties(T)
 end

module PDProps = DTreeProps(PTree)

module ADProps = DTreeProps(ATree)

module DTreeConv = 
 struct 
  (** val sdtree_tdtree :
      ATree.elt PTree.t -> PTree.elt coq_DTree -> ATree.elt coq_DTree option **)
  
  let sdtree_tdtree s2t =
    let rec sdtree_tdtree0 = function
      | DT_node (s, sdts) ->
          (match PTree.get s s2t with
             | Some t0 ->
                 (match sdtrees_tdtrees0 sdts with
                    | Some tdts -> Some (DT_node (t0, tdts))
                    | None -> None)
             | None -> None)
    and sdtrees_tdtrees0 = function
      | DT_nil -> Some DT_nil
      | DT_cons (sdt, sdts0) ->
          (match sdtree_tdtree0 sdt with
             | Some tdt ->
                 (match sdtrees_tdtrees0 sdts0 with
                    | Some tdts -> Some (DT_cons (tdt, tdts))
                    | None -> None)
             | None -> None)
    in sdtree_tdtree0
  
  (** val sdtrees_tdtrees :
      ATree.elt PTree.t -> PTree.elt coq_DTrees -> ATree.elt coq_DTrees
      option **)
  
  let sdtrees_tdtrees s2t =
    let rec sdtree_tdtree0 = function
      | DT_node (s, sdts) ->
          (match PTree.get s s2t with
             | Some t0 ->
                 (match sdtrees_tdtrees0 sdts with
                    | Some tdts -> Some (DT_node (t0, tdts))
                    | None -> None)
             | None -> None)
    and sdtrees_tdtrees0 = function
      | DT_nil -> Some DT_nil
      | DT_cons (sdt, sdts0) ->
          (match sdtree_tdtree0 sdt with
             | Some tdt ->
                 (match sdtrees_tdtrees0 sdts0 with
                    | Some tdts -> Some (DT_cons (tdt, tdts))
                    | None -> None)
             | None -> None)
    in sdtrees_tdtrees0
 end

