open Datatypes
open List0
open ListSet
open Maps

type 'a coq_DTree =
  | DT_node of 'a * 'a coq_DTrees
and 'a coq_DTrees =
  | DT_nil
  | DT_cons of 'a coq_DTree * 'a coq_DTrees

val coq_DTree_rect : ('a1 -> 'a1 coq_DTrees -> 'a2) -> 'a1 coq_DTree -> 'a2

val coq_DTree_rec : ('a1 -> 'a1 coq_DTrees -> 'a2) -> 'a1 coq_DTree -> 'a2

val coq_DTrees_rect :
  'a2 -> ('a1 coq_DTree -> 'a1 coq_DTrees -> 'a2 -> 'a2) -> 'a1 coq_DTrees ->
  'a2

val coq_DTrees_rec :
  'a2 -> ('a1 coq_DTree -> 'a1 coq_DTrees -> 'a2 -> 'a2) -> 'a1 coq_DTrees ->
  'a2

val size_of_dtrees : 'a1 coq_DTrees -> nat

val list2tree : 'a1 -> 'a1 list -> 'a1 coq_DTree

val dtree_insert :
  ('a1 -> 'a1 -> bool) -> 'a1 coq_DTree -> 'a1 -> 'a1 -> 'a1 list -> 'a1
  coq_DTree

val dtrees_insert :
  ('a1 -> 'a1 -> bool) -> 'a1 coq_DTrees -> 'a1 -> 'a1 list -> 'a1 coq_DTrees

val create_dtree_from_chain :
  ('a1 -> 'a1 -> bool) -> 'a1 coq_DTree -> 'a1 list -> 'a1 coq_DTree

val create_dtree_from_chains :
  ('a1 -> 'a1 -> bool) -> 'a1 -> ('a1*'a1 list) list -> 'a1 coq_DTree

val in_children_roots : ('a1 -> 'a1 -> bool) -> 'a1 -> 'a1 coq_DTrees -> bool

val is_dtree_edge :
  ('a1 -> 'a1 -> bool) -> 'a1 coq_DTree -> 'a1 -> 'a1 -> bool

val is_dtrees_edge :
  ('a1 -> 'a1 -> bool) -> 'a1 coq_DTrees -> 'a1 -> 'a1 -> bool

val dtrees_rec2 :
  ('a1 -> 'a1 coq_DTrees -> 'a3 -> 'a2) -> 'a3 -> ('a1 coq_DTree -> 'a2 ->
  'a1 coq_DTrees -> 'a3 -> 'a3) -> 'a1 coq_DTrees -> 'a3

val dtree_rec2 :
  ('a1 -> 'a1 coq_DTrees -> 'a3 -> 'a2) -> 'a3 -> ('a1 coq_DTree -> 'a2 ->
  'a1 coq_DTrees -> 'a3 -> 'a3) -> 'a1 coq_DTree -> 'a2

val dtree_mutrec :
  ('a1 -> 'a1 coq_DTrees -> 'a3 -> 'a2) -> 'a3 -> ('a1 coq_DTree -> 'a2 ->
  'a1 coq_DTrees -> 'a3 -> 'a3) -> ('a1 coq_DTree -> 'a2)*('a1 coq_DTrees ->
  'a3)

val get_dtree_root : 'a1 coq_DTree -> 'a1

val dtree_dom : ('a1 -> 'a1 -> bool) -> 'a1 coq_DTree -> 'a1 set

val dtrees_dom : ('a1 -> 'a1 -> bool) -> 'a1 coq_DTrees -> 'a1 set

val get_dtrees_roots : ('a1 -> 'a1 -> bool) -> 'a1 coq_DTrees -> 'a1 set

module DTreeProps : 
 functor (T:TREE) ->
 sig 
  module TCfg : 
   sig 
    module XTree : 
     sig 
      module TProp : 
       sig 
        val f' : ('a2 -> T.elt -> 'a1 -> 'a2) -> 'a2 -> (T.elt*'a1) -> 'a2
       end
      
      val successors_list : 'a1 list T.t -> T.elt -> 'a1 list
      
      val add_successors :
        T.elt list T.t -> T.elt -> T.elt list -> T.elt list T.t
      
      val make_predecessors : T.elt list T.t -> T.elt list T.t
      
      val children_of_tree : 'a1 list T.t -> 'a1 list
      
      val parents_of_tree : 'a1 T.t -> T.elt list
      
      val elements_of_cfg :
        T.elt list T.t -> (T.elt -> T.elt -> bool) -> T.elt list
     end
   end
  
  module XTree : 
   sig 
    module TProp : 
     sig 
      val f' : ('a2 -> T.elt -> 'a1 -> 'a2) -> 'a2 -> (T.elt*'a1) -> 'a2
     end
    
    val successors_list : 'a1 list T.t -> T.elt -> 'a1 list
    
    val add_successors :
      T.elt list T.t -> T.elt -> T.elt list -> T.elt list T.t
    
    val make_predecessors : T.elt list T.t -> T.elt list T.t
    
    val children_of_tree : 'a1 list T.t -> 'a1 list
    
    val parents_of_tree : 'a1 T.t -> T.elt list
    
    val elements_of_cfg :
      T.elt list T.t -> (T.elt -> T.elt -> bool) -> T.elt list
   end
 end

module PDProps : 
 sig 
  module TCfg : 
   sig 
    module XTree : 
     sig 
      module TProp : 
       sig 
        val f' :
          ('a2 -> PTree.elt -> 'a1 -> 'a2) -> 'a2 -> (PTree.elt*'a1) -> 'a2
       end
      
      val successors_list : 'a1 list PTree.t -> PTree.elt -> 'a1 list
      
      val add_successors :
        PTree.elt list PTree.t -> PTree.elt -> PTree.elt list -> PTree.elt
        list PTree.t
      
      val make_predecessors :
        PTree.elt list PTree.t -> PTree.elt list PTree.t
      
      val children_of_tree : 'a1 list PTree.t -> 'a1 list
      
      val parents_of_tree : 'a1 PTree.t -> PTree.elt list
      
      val elements_of_cfg :
        PTree.elt list PTree.t -> (PTree.elt -> PTree.elt -> bool) ->
        PTree.elt list
     end
   end
  
  module XTree : 
   sig 
    module TProp : 
     sig 
      val f' :
        ('a2 -> PTree.elt -> 'a1 -> 'a2) -> 'a2 -> (PTree.elt*'a1) -> 'a2
     end
    
    val successors_list : 'a1 list PTree.t -> PTree.elt -> 'a1 list
    
    val add_successors :
      PTree.elt list PTree.t -> PTree.elt -> PTree.elt list -> PTree.elt list
      PTree.t
    
    val make_predecessors : PTree.elt list PTree.t -> PTree.elt list PTree.t
    
    val children_of_tree : 'a1 list PTree.t -> 'a1 list
    
    val parents_of_tree : 'a1 PTree.t -> PTree.elt list
    
    val elements_of_cfg :
      PTree.elt list PTree.t -> (PTree.elt -> PTree.elt -> bool) -> PTree.elt
      list
   end
 end

module ADProps : 
 sig 
  module TCfg : 
   sig 
    module XTree : 
     sig 
      module TProp : 
       sig 
        val f' :
          ('a2 -> ATree.elt -> 'a1 -> 'a2) -> 'a2 -> (ATree.elt*'a1) -> 'a2
       end
      
      val successors_list : 'a1 list ATree.t -> ATree.elt -> 'a1 list
      
      val add_successors :
        ATree.elt list ATree.t -> ATree.elt -> ATree.elt list -> ATree.elt
        list ATree.t
      
      val make_predecessors :
        ATree.elt list ATree.t -> ATree.elt list ATree.t
      
      val children_of_tree : 'a1 list ATree.t -> 'a1 list
      
      val parents_of_tree : 'a1 ATree.t -> ATree.elt list
      
      val elements_of_cfg :
        ATree.elt list ATree.t -> (ATree.elt -> ATree.elt -> bool) ->
        ATree.elt list
     end
   end
  
  module XTree : 
   sig 
    module TProp : 
     sig 
      val f' :
        ('a2 -> ATree.elt -> 'a1 -> 'a2) -> 'a2 -> (ATree.elt*'a1) -> 'a2
     end
    
    val successors_list : 'a1 list ATree.t -> ATree.elt -> 'a1 list
    
    val add_successors :
      ATree.elt list ATree.t -> ATree.elt -> ATree.elt list -> ATree.elt list
      ATree.t
    
    val make_predecessors : ATree.elt list ATree.t -> ATree.elt list ATree.t
    
    val children_of_tree : 'a1 list ATree.t -> 'a1 list
    
    val parents_of_tree : 'a1 ATree.t -> ATree.elt list
    
    val elements_of_cfg :
      ATree.elt list ATree.t -> (ATree.elt -> ATree.elt -> bool) -> ATree.elt
      list
   end
 end

module DTreeConv : 
 sig 
  val sdtree_tdtree :
    ATree.elt PTree.t -> PTree.elt coq_DTree -> ATree.elt coq_DTree option
  
  val sdtrees_tdtrees :
    ATree.elt PTree.t -> PTree.elt coq_DTrees -> ATree.elt coq_DTrees option
 end

