open BinPos
open Datatypes
open Iteration
open List0
open Maps
open Dfs
open Dom_libs
open Infrastructure
open Syntax

module Weak_Pred_Dataflow_Solver = 
 functor (NS:PNODE_SET) ->
 functor (L:LATTICEELT) ->
 struct 
  type state = { st_in : L.t PMap.t; st_wrk : NS.t }
  
  (** val state_rect : (L.t PMap.t -> NS.t -> 'a1) -> state -> 'a1 **)
  
  let state_rect f s =
    let { st_in = x; st_wrk = x0 } = s in f x x0
  
  (** val state_rec : (L.t PMap.t -> NS.t -> 'a1) -> state -> 'a1 **)
  
  let state_rec f s =
    let { st_in = x; st_wrk = x0 } = s in f x x0
  
  (** val st_in : state -> L.t PMap.t **)
  
  let st_in s =
    s.st_in
  
  (** val st_wrk : state -> NS.t **)
  
  let st_wrk s =
    s.st_wrk
  
  (** val start_state_in : (positive*L.t) list -> L.t PMap.t **)
  
  let rec start_state_in = function
    | [] -> PMap.init L.bot
    | p::rem ->
        let n,v = p in
        let m = start_state_in rem in
        PMap.set n (fst (L.lub (PMap.get n m) v)) m
  
  (** val start_state :
      positive list PTree.t -> (positive*L.t) list -> state **)
  
  let start_state predecessors entrypoints =
    { st_in = (start_state_in entrypoints); st_wrk =
      (NS.initial predecessors) }
  
  (** val propagate_pred_list :
      (positive -> L.t -> L.t) -> state -> L.t -> positive list -> L.t*bool **)
  
  let propagate_pred_list transf s oldl preds =
    fold_left (fun acc p ->
      let accl,accb = acc in
      let newl,changed = L.lub accl (transf p (PMap.get p (st_in s))) in
      newl,(if accb then true else changed)) preds (oldl,false)
  
  (** val add_successors_into_worklist :
      positive list PTree.t -> NS.t -> positive -> NS.t **)
  
  let add_successors_into_worklist successors0 rem n =
    fold_left (fun acc s -> NS.add s acc)
      (XPTree.successors_list successors0 n) rem
  
  (** val step :
      positive list PTree.t -> positive list PTree.t -> (positive -> L.t ->
      L.t) -> state -> (L.t PMap.t, state) sum **)
  
  let step successors0 predecessors transf s =
    match NS.pick (st_wrk s) with
      | Some p ->
          let n,rem = p in
          let oldl = PMap.get n (st_in s) in
          let newl,changed =
            propagate_pred_list transf s oldl
              (XPTree.successors_list predecessors n)
          in
          let s' =
            if changed
            then { st_in = (PMap.set n newl (st_in s)); st_wrk =
                   (add_successors_into_worklist successors0 rem n) }
            else { st_in = (st_in s); st_wrk = rem }
          in
          Coq_inr s'
      | None -> Coq_inl (st_in s)
  
  (** val fixpoint :
      positive list PTree.t -> positive list PTree.t -> (positive -> L.t ->
      L.t) -> (positive*L.t) list -> L.t PMap.t option **)
  
  let fixpoint successors0 predecessors transf entrypoints =
    PrimIter.iterate (step successors0 predecessors transf)
      (start_state predecessors entrypoints)
 end

module LDoms = Doms(MergeLt)

module DomDS = Weak_Pred_Dataflow_Solver(PNodeSetMax)(LDoms)

(** val transfer : positive -> LDoms.t -> LDoms.t **)

let transfer n = function
  | Some ps -> Some (n::ps)
  | None -> None

(** val dom_analyze : LLVMsyntax.fdef -> LDoms.t PMap.t **)

let dom_analyze f =
  let asuccs = Cfg.successors f in
  (match LLVMinfra.getEntryBlock f with
     | Some b ->
         let LLVMsyntax.Coq_block_intro (le, phinodes5, cmds5, terminator5) =
           b
         in
         let { coq_PO_cnt = pO_cnt; coq_PO_a2p = a2p } = dfs asuccs le Coq_xH
         in
         let psuccs = asuccs_psuccs a2p asuccs in
         (match ATree.get le a2p with
            | Some pe ->
                (match DomDS.fixpoint psuccs
                         (XPTree.make_predecessors psuccs) transfer
                         ((pe,LDoms.top)::[]) with
                   | Some res -> res
                   | None -> PMap.set pe LDoms.top (PMap.init LDoms.bot))
            | None -> PMap.init LDoms.bot)
     | None -> PMap.init LDoms.bot)

