open BinPos
open Datatypes
open Iteration
open List0
open Maps
open Dfs
open Dom_libs
open Infrastructure
open Syntax

module Weak_Pred_Dataflow_Solver : 
 functor (NS:PNODE_SET) ->
 functor (L:LATTICEELT) ->
 sig 
  type state = { st_in : L.t PMap.t; st_wrk : NS.t }
  
  val state_rect : (L.t PMap.t -> NS.t -> 'a1) -> state -> 'a1
  
  val state_rec : (L.t PMap.t -> NS.t -> 'a1) -> state -> 'a1
  
  val st_in : state -> L.t PMap.t
  
  val st_wrk : state -> NS.t
  
  val start_state_in : (positive*L.t) list -> L.t PMap.t
  
  val start_state : positive list PTree.t -> (positive*L.t) list -> state
  
  val propagate_pred_list :
    (positive -> L.t -> L.t) -> state -> L.t -> positive list -> L.t*bool
  
  val add_successors_into_worklist :
    positive list PTree.t -> NS.t -> positive -> NS.t
  
  val step :
    positive list PTree.t -> positive list PTree.t -> (positive -> L.t ->
    L.t) -> state -> (L.t PMap.t, state) sum
  
  val fixpoint :
    positive list PTree.t -> positive list PTree.t -> (positive -> L.t ->
    L.t) -> (positive*L.t) list -> L.t PMap.t option
 end

module LDoms : 
 sig 
  type t = positive list option
  
  val bot : t
  
  val lub : t -> t -> t*bool
  
  val top : t
  
  val transfer : positive -> t -> t
 end

module DomDS : 
 sig 
  type state = { st_in : LDoms.t PMap.t; st_wrk : PNodeSetMax.t }
  
  val state_rect : (LDoms.t PMap.t -> PNodeSetMax.t -> 'a1) -> state -> 'a1
  
  val state_rec : (LDoms.t PMap.t -> PNodeSetMax.t -> 'a1) -> state -> 'a1
  
  val st_in : state -> LDoms.t PMap.t
  
  val st_wrk : state -> PNodeSetMax.t
  
  val start_state_in : (positive*LDoms.t) list -> LDoms.t PMap.t
  
  val start_state : positive list PTree.t -> (positive*LDoms.t) list -> state
  
  val propagate_pred_list :
    (positive -> LDoms.t -> LDoms.t) -> state -> LDoms.t -> positive list ->
    LDoms.t*bool
  
  val add_successors_into_worklist :
    positive list PTree.t -> PNodeSetMax.t -> positive -> PNodeSetMax.t
  
  val step :
    positive list PTree.t -> positive list PTree.t -> (positive -> LDoms.t ->
    LDoms.t) -> state -> (LDoms.t PMap.t, state) sum
  
  val fixpoint :
    positive list PTree.t -> positive list PTree.t -> (positive -> LDoms.t ->
    LDoms.t) -> (positive*LDoms.t) list -> LDoms.t PMap.t option
 end

val transfer : positive -> LDoms.t -> LDoms.t

val dom_analyze : LLVMsyntax.fdef -> LDoms.t PMap.t

