open BinPos
open Datatypes
open Iteration
open List0
open Maps
open Dom_libs

module Weak_Succ_Dataflow_Solver = 
 functor (NS:PNODE_SET) ->
 functor (L:LATTICEELT) ->
 struct 
  type state = { st_in : L.t PMap.t; st_wrk : NS.t }
  
  (** val state_rect : (L.t PMap.t -> NS.t -> 'a1) -> state -> 'a1 **)
  
  let state_rect f s =
    let { st_in = x; st_wrk = x0 } = s in f x x0
  
  (** val state_rec : (L.t PMap.t -> NS.t -> 'a1) -> state -> 'a1 **)
  
  let state_rec f s =
    let { st_in = x; st_wrk = x0 } = s in f x x0
  
  (** val st_in : state -> L.t PMap.t **)
  
  let st_in s =
    s.st_in
  
  (** val st_wrk : state -> NS.t **)
  
  let st_wrk s =
    s.st_wrk
  
  (** val start_state_in : (positive*L.t) list -> L.t PMap.t **)
  
  let rec start_state_in = function
    | [] -> PMap.init L.bot
    | p::rem ->
        let n,v = p in
        let m = start_state_in rem in
        PMap.set n (fst (L.lub (PMap.get n m) v)) m
  
  (** val start_state :
      positive list PTree.t -> (positive*L.t) list -> state **)
  
  let start_state successors entrypoints0 =
    { st_in = (start_state_in entrypoints0); st_wrk =
      (NS.initial successors) }
  
  (** val propagate_succ : state -> L.t -> positive -> state **)
  
  let propagate_succ s out n =
    let oldl = PMap.get n (st_in s) in
    let newl,changed = L.lub oldl out in
    if changed
    then { st_in = (PMap.set n newl (st_in s)); st_wrk =
           (NS.add n (st_wrk s)) }
    else s
  
  (** val propagate_succ_list : state -> L.t -> positive list -> state **)
  
  let rec propagate_succ_list s out = function
    | [] -> s
    | n::rem -> propagate_succ_list (propagate_succ s out n) out rem
  
  (** val step :
      positive list PTree.t -> (positive -> L.t -> L.t) -> state -> (L.t
      PMap.t, state) sum **)
  
  let step successors transf s =
    match NS.pick (st_wrk s) with
      | Some p ->
          let n,rem = p in
          Coq_inr
          (propagate_succ_list { st_in = (st_in s); st_wrk = rem }
            (transf n (PMap.get n (st_in s)))
            (XPTree.successors_list successors n))
      | None -> Coq_inl (st_in s)
  
  (** val fixpoint :
      positive list PTree.t -> (positive -> L.t -> L.t) -> (positive*L.t)
      list -> positive -> L.t PMap.t option **)
  
  let fixpoint successors transf entrypoints0 num_iterations =
    PrimIter.iter (step successors transf) num_iterations
      (start_state successors entrypoints0)
 end

module LDoms = Doms(MergeLt)

module DomDS = Weak_Succ_Dataflow_Solver(PNodeSetMax)(LDoms)

module DomMap = LATTICEELT_MAP(LDoms)

module WorklistProps = 
 struct 
  
 end

module InitOrder = 
 struct 
  (** val predecessors : positive list PTree.t -> PTree.elt list PTree.t **)
  
  let predecessors successors =
    XPTree.make_predecessors successors
  
  (** val entrypoints : positive -> (positive*LDoms.t) list **)
  
  let entrypoints entrypoint =
    (entrypoint,LDoms.top)::[]
 end

module Mono = 
 struct 
  
 end

module DomsInParents = 
 struct 
  (** val entrypoints : positive -> (positive*LDoms.t) list **)
  
  let entrypoints entrypoint =
    (entrypoint,LDoms.top)::[]
 end

module NonCFGIsBot = 
 struct 
  (** val entrypoints : positive -> LDoms.t -> (positive*LDoms.t) list **)
  
  let entrypoints entrypoint init0 =
    (entrypoint,init0)::[]
 end

module LtDoms = 
 struct 
  (** val entrypoints : positive -> (positive*LDoms.t) list **)
  
  let entrypoints entrypoint =
    (entrypoint,LDoms.top)::[]
  
  (** val predecessors : positive list PTree.t -> PTree.elt list PTree.t **)
  
  let predecessors successors =
    XPTree.make_predecessors successors
 end

module SortedDoms = 
 struct 
  (** val entrypoints : positive -> (positive*LDoms.t) list **)
  
  let entrypoints entrypoint =
    (entrypoint,LDoms.top)::[]
  
  (** val predecessors : positive list PTree.t -> PTree.elt list PTree.t **)
  
  let predecessors successors =
    XPTree.make_predecessors successors
 end

module Inequations = 
 struct 
  (** val entrypoints : positive -> (positive*LDoms.t) list **)
  
  let entrypoints entrypoint =
    (entrypoint,LDoms.top)::[]
  
  (** val predecessors : positive list PTree.t -> PTree.elt list PTree.t **)
  
  let predecessors successors =
    XPTree.make_predecessors successors
 end

(** val pdom_analyze :
    positive list PTree.t -> positive -> positive -> LDoms.t PMap.t **)

let pdom_analyze successors entrypoint num_iters0 =
  match DomDS.fixpoint successors LDoms.transfer ((entrypoint,LDoms.top)::[])
          num_iters0 with
    | Some res -> res
    | None -> PMap.init LDoms.top

(** val entrypoints : positive -> (positive*LDoms.t) list **)

let entrypoints entrypoint =
  (entrypoint,LDoms.top)::[]

module Termination = 
 struct 
  (** val elements_of_cfg : positive list PTree.t -> positive list **)
  
  let elements_of_cfg successors =
    XPTree.elements_of_cfg successors positive_eq_dec
  
  (** val psize_of_cfg : positive list PTree.t -> positive **)
  
  let psize_of_cfg successors =
    coq_Pplus (Cfg.plength (elements_of_cfg successors)) (Coq_xO Coq_xH)
  
  (** val num_of_doms_fun :
      positive list PTree.t -> LDoms.t PMap.t -> positive -> positive ->
      positive **)
  
  let num_of_doms_fun successors dmap acc p =
    coq_Pplus acc
      (match PMap.get p dmap with
         | Some x -> Cfg.plength x
         | None -> psize_of_cfg successors)
  
  (** val num_of_doms :
      positive list PTree.t -> LDoms.t PMap.t -> positive **)
  
  let num_of_doms successors dmap =
    fold_left (num_of_doms_fun successors dmap) (elements_of_cfg successors)
      Coq_xH
  
  (** val num_iters : positive list PTree.t -> positive **)
  
  let num_iters successors =
    Cfg.coq_Pcubeplus (psize_of_cfg successors)
  
  (** val psize_of_worklist : PNodeSetMax.t -> positive **)
  
  let psize_of_worklist wrk =
    coq_Pplus (coq_P_of_succ_nat (PositiveSet.cardinal wrk)) Coq_xH
  
  (** val num_iters_aux :
      positive list PTree.t -> DomDS.state -> positive **)
  
  let num_iters_aux successors st =
    coq_Pplus
      (coq_Pmult (psize_of_cfg successors)
        (num_of_doms successors st.DomDS.st_in))
      (psize_of_worklist st.DomDS.st_wrk)
  
  (** val entrypoints : positive -> (positive*LDoms.t) list **)
  
  let entrypoints entrypoint =
    (entrypoint,LDoms.top)::[]
  
  (** val predecessors : positive list PTree.t -> PTree.elt list PTree.t **)
  
  let predecessors successors =
    XPTree.make_predecessors successors
 end

module UnreachableDoms = 
 struct 
  (** val entrypoints : positive -> (positive*LDoms.t) list **)
  
  let entrypoints entrypoint =
    (entrypoint,LDoms.top)::[]
 end

module EntryDomsOthers = 
 struct 
  (** val entrypoints : positive -> (positive*LDoms.t) list **)
  
  let entrypoints entrypoint =
    (entrypoint,LDoms.top)::[]
  
  (** val predecessors : positive list PTree.t -> PTree.elt list PTree.t **)
  
  let predecessors successors =
    XPTree.make_predecessors successors
 end

module DomComplete = 
 struct 
  (** val entrypoints : positive -> (positive*LDoms.t) list **)
  
  let entrypoints entrypoint =
    (entrypoint,LDoms.top)::[]
  
  (** val predecessors : positive list PTree.t -> PTree.elt list PTree.t **)
  
  let predecessors successors =
    XPTree.make_predecessors successors
 end

module DomSound = 
 struct 
  (** val entrypoints : positive -> (positive*LDoms.t) list **)
  
  let entrypoints entrypoint =
    (entrypoint,LDoms.top)::[]
  
  (** val predecessors : positive list PTree.t -> PTree.elt list PTree.t **)
  
  let predecessors successors =
    XPTree.make_predecessors successors
 end

module PDomProps = 
 struct 
  (** val entrypoints : positive -> (positive*LDoms.t) list **)
  
  let entrypoints entrypoint =
    (entrypoint,LDoms.top)::[]
  
  (** val predecessors : positive list PTree.t -> PTree.elt list PTree.t **)
  
  let predecessors successors =
    XPTree.make_predecessors successors
 end

module IdomSorted = 
 struct 
  (** val entrypoints : positive -> (positive*LDoms.t) list **)
  
  let entrypoints entrypoint =
    (entrypoint,LDoms.top)::[]
  
  (** val predecessors : positive list PTree.t -> PTree.elt list PTree.t **)
  
  let predecessors successors =
    XPTree.make_predecessors successors
 end

