open BinPos
open Datatypes
open Iteration
open List0
open Maps
open Dom_libs

module Weak_Succ_Dataflow_Solver : 
 functor (NS:PNODE_SET) ->
 functor (L:LATTICEELT) ->
 sig 
  type state = { st_in : L.t PMap.t; st_wrk : NS.t }
  
  val state_rect : (L.t PMap.t -> NS.t -> 'a1) -> state -> 'a1
  
  val state_rec : (L.t PMap.t -> NS.t -> 'a1) -> state -> 'a1
  
  val st_in : state -> L.t PMap.t
  
  val st_wrk : state -> NS.t
  
  val start_state_in : (positive*L.t) list -> L.t PMap.t
  
  val start_state : positive list PTree.t -> (positive*L.t) list -> state
  
  val propagate_succ : state -> L.t -> positive -> state
  
  val propagate_succ_list : state -> L.t -> positive list -> state
  
  val step :
    positive list PTree.t -> (positive -> L.t -> L.t) -> state -> (L.t
    PMap.t, state) sum
  
  val fixpoint :
    positive list PTree.t -> (positive -> L.t -> L.t) -> (positive*L.t) list
    -> positive -> L.t PMap.t option
 end

module LDoms : 
 sig 
  type t = positive list option
  
  val bot : t
  
  val lub : t -> t -> t*bool
  
  val top : t
  
  val transfer : positive -> t -> t
 end

module DomDS : 
 sig 
  type state = { st_in : LDoms.t PMap.t; st_wrk : PNodeSetMax.t }
  
  val state_rect : (LDoms.t PMap.t -> PNodeSetMax.t -> 'a1) -> state -> 'a1
  
  val state_rec : (LDoms.t PMap.t -> PNodeSetMax.t -> 'a1) -> state -> 'a1
  
  val st_in : state -> LDoms.t PMap.t
  
  val st_wrk : state -> PNodeSetMax.t
  
  val start_state_in : (positive*LDoms.t) list -> LDoms.t PMap.t
  
  val start_state : positive list PTree.t -> (positive*LDoms.t) list -> state
  
  val propagate_succ : state -> LDoms.t -> positive -> state
  
  val propagate_succ_list : state -> LDoms.t -> positive list -> state
  
  val step :
    positive list PTree.t -> (positive -> LDoms.t -> LDoms.t) -> state ->
    (LDoms.t PMap.t, state) sum
  
  val fixpoint :
    positive list PTree.t -> (positive -> LDoms.t -> LDoms.t) ->
    (positive*LDoms.t) list -> positive -> LDoms.t PMap.t option
 end

module DomMap : 
 sig 
  
 end

module WorklistProps : 
 sig 
  
 end

module InitOrder : 
 sig 
  val predecessors : positive list PTree.t -> PTree.elt list PTree.t
  
  val entrypoints : positive -> (positive*LDoms.t) list
 end

module Mono : 
 sig 
  
 end

module DomsInParents : 
 sig 
  val entrypoints : positive -> (positive*LDoms.t) list
 end

module NonCFGIsBot : 
 sig 
  val entrypoints : positive -> LDoms.t -> (positive*LDoms.t) list
 end

module LtDoms : 
 sig 
  val entrypoints : positive -> (positive*LDoms.t) list
  
  val predecessors : positive list PTree.t -> PTree.elt list PTree.t
 end

module SortedDoms : 
 sig 
  val entrypoints : positive -> (positive*LDoms.t) list
  
  val predecessors : positive list PTree.t -> PTree.elt list PTree.t
 end

module Inequations : 
 sig 
  val entrypoints : positive -> (positive*LDoms.t) list
  
  val predecessors : positive list PTree.t -> PTree.elt list PTree.t
 end

val pdom_analyze :
  positive list PTree.t -> positive -> positive -> LDoms.t PMap.t

val entrypoints : positive -> (positive*LDoms.t) list

module Termination : 
 sig 
  val elements_of_cfg : positive list PTree.t -> positive list
  
  val psize_of_cfg : positive list PTree.t -> positive
  
  val num_of_doms_fun :
    positive list PTree.t -> LDoms.t PMap.t -> positive -> positive ->
    positive
  
  val num_of_doms : positive list PTree.t -> LDoms.t PMap.t -> positive
  
  val num_iters : positive list PTree.t -> positive
  
  val psize_of_worklist : PNodeSetMax.t -> positive
  
  val num_iters_aux : positive list PTree.t -> DomDS.state -> positive
  
  val entrypoints : positive -> (positive*LDoms.t) list
  
  val predecessors : positive list PTree.t -> PTree.elt list PTree.t
 end

module UnreachableDoms : 
 sig 
  val entrypoints : positive -> (positive*LDoms.t) list
 end

module EntryDomsOthers : 
 sig 
  val entrypoints : positive -> (positive*LDoms.t) list
  
  val predecessors : positive list PTree.t -> PTree.elt list PTree.t
 end

module DomComplete : 
 sig 
  val entrypoints : positive -> (positive*LDoms.t) list
  
  val predecessors : positive list PTree.t -> PTree.elt list PTree.t
 end

module DomSound : 
 sig 
  val entrypoints : positive -> (positive*LDoms.t) list
  
  val predecessors : positive list PTree.t -> PTree.elt list PTree.t
 end

module PDomProps : 
 sig 
  val entrypoints : positive -> (positive*LDoms.t) list
  
  val predecessors : positive list PTree.t -> PTree.elt list PTree.t
 end

module IdomSorted : 
 sig 
  val entrypoints : positive -> (positive*LDoms.t) list
  
  val predecessors : positive list PTree.t -> PTree.elt list PTree.t
 end

